from typing import List, Dict, Union, Optional, Set
from uuid import UUID
from pyarrow import csv, Table

from datalogue.clients.http import _HttpClient, HttpMethod, MultiPartData, MultiPartFile
from datalogue.models.schema import AbstractDataSchema, Schema, SchemaNode
from datalogue.models.datastore import Datastore, _datastore_from_payload, FileFormat, Cell, _cell_from_payload, \
    NodeClassification
from datalogue.errors import DtlError
from datalogue.dtl_utils import _parse_list, ResponseStream
from datalogue.anduin.client import _AnduinClient
from datalogue.clients.ontology import _OntologyClient
from datalogue.anduin.models.stream import StreamStatus
from functools import reduce


class _DatastoreClient:
    """
    Client to interact with the Datastores objects
    """

    def __init__(self, http_client: _HttpClient):
        self.http_client = http_client
        self.service_uri = "/scout"

    def create(self, datastore: Datastore) -> Union[DtlError, Datastore]:
        """
        Creates a data source

        :param datastore: data source to be created
        :return: string with error message if failed, the data source otherwise
        """
        res = self.http_client.make_authed_request(
            self.service_uri + "/datastores", HttpMethod.POST, datastore._as_payload())

        if isinstance(res, DtlError):
            return res

        return _datastore_from_payload(res)

    def get(self, datastore_id: UUID) -> Union[DtlError, Datastore]:
        """
        From the provided id, get the corresponding Datastore

        :param datastore_id: id of the datastore to be retrieved
        :return:
        """

        res = self.http_client.make_authed_request(self.service_uri + "/datastores/" + str(datastore_id),
                                                   HttpMethod.GET)

        if isinstance(res, DtlError):
            return res

        datastore_resp = _datastore_from_payload(res)
        if isinstance(datastore_resp, DtlError):
            return datastore_resp

        return self.__load_datastore_with_class_info(datastore_resp)

    def update(self, datastore: Datastore) -> Union[DtlError, Datastore]:
        """
        Updates the backend with the new status of the existing datastore

        :param datastore: to be persisted
        :return:
        """

        res = self.http_client.make_authed_request(
            self.service_uri + "/datastores/" + str(datastore.id),
            HttpMethod.PUT,
            datastore._as_payload()
        )

        if isinstance(res, DtlError):
            return res

        return _datastore_from_payload(res)

    def list(self, page: int = 1, item_per_page: int = 25) -> Union[DtlError, List[Datastore]]:
        """
        List all the datastores that are saved

        TODO pagination

        :param page: page to be retrieved (ignored)
        :param item_per_page: number of items to be put in a page (ignored)
        :return: Returns a List of all the available datastores or an error message as a string
        """

        res = self.http_client.make_authed_request(self.service_uri + "/datastores", HttpMethod.GET)

        if isinstance(res, DtlError):
            return res

        datastore_list = _parse_list(_datastore_from_payload)(res)
        if isinstance(datastore_list, DtlError):
            return datastore_list

        datastore_list_with_class_info = []
        for datastore_resp in datastore_list:
            datastore_resp_with_class_info = self.__load_datastore_with_class_info(datastore_resp)
            if isinstance(datastore_resp_with_class_info, DtlError):
                return datastore_resp_with_class_info
            datastore_list_with_class_info.append(datastore_resp_with_class_info)
        return datastore_list_with_class_info

    def search(self,
               by_name: Optional[List[str]] = None,
               by_field_name: Optional[List[str]] = None,
               by_field_class: Optional[List[str]] = None,
               page: int = 1,
               item_per_page: int = 25) -> Union[DtlError, List[Datastore]]:
        """
        Search among all backend Datastores, using any numbers of queries as the search criteria
        Multiple items within a query will be applied with OR logic — multiple queries will be applied with AND logic

        The wildcard operators: `?` and `*` can be used as wildcards,
        matching one and any number of additional characters, respectively

        :param by_name: a list of datastore names
        :param by_field_name: a list of column labels
        :param by_field_class: a list of classes that columns are classified
        :param page: page to be retrieved
        :param item_per_page: number of items to be put in a page
        :return: Returns a List of all the available datastores or an error message as a string
        """

        queries = {}

        if isinstance(by_name, list):
            queries["name"] = by_name

        if isinstance(by_field_name, list):
            queries["schema.nodes.label"] = by_field_name

        if isinstance(by_field_class, list):
            queries["classes.name"] = by_field_class

        if len(queries) > 0:
            payload = {
                "queries": queries,
                "page": page,
                "size": item_per_page,
                "type": "datastore"
            }

            res = self.http_client.make_authed_request(self.service_uri + "/v2/search", HttpMethod.POST, payload)
            if isinstance(res, DtlError):
                return res
            return _parse_list(_datastore_from_payload)(res['results'])
        else:
            # If there is not query, we will bring everything
            payload = {
                "query": "*",
                "page": page,
                "size": item_per_page,
                "type": "datastore"
            }

            res = self.http_client.make_authed_request(self.service_uri + "/search", HttpMethod.POST, payload)
            return _parse_list(_datastore_from_payload)(res['results'])

    def get_schema(self, datastore_id: UUID) -> Union[DtlError, AbstractDataSchema]:
        """
        From the provided id, get the corresponding Datastore schema

        :param datastore_id: id of the datastore to be retrieved
        :return:
        """
        res = self.http_client.make_authed_request(self.service_uri + "/datastores/" + str(datastore_id) + "/schema",
                                                   HttpMethod.GET)
        if isinstance(res, DtlError):
            return res

        schema = Schema._from_payload(res).data_schema
        if isinstance(schema, DtlError):
            return schema

        schema = self.__load_ads_with_class_info(schema)
        return schema

    def __load_datastore_with_class_info(self, datastore: Datastore) -> Union[DtlError, Datastore]:
        if datastore.ads is None and (datastore.schema_nodes is None or len(datastore.schema_nodes) is 0):
            return datastore

        if datastore.ads is not None or datastore.schema_nodes is not None:
            # 1. Get schema nodes from either Datastore or from ADS
            schema_nodes = []
            if datastore.schema_nodes is not None:
                schema_nodes = datastore.schema_nodes
            elif datastore.ads is not None:
                schema_nodes = list(map(lambda node_def: node_def.node, datastore.ads.get_nodes()))

            # 2. Load the dictionary for set of class_id
            class_id_path_dict = self.__load_paths_of_class_ids(schema_nodes)
            if isinstance(class_id_path_dict, DtlError):
                return class_id_path_dict

            # 3. Update ADS and List of SchemaNode respectively
            if datastore.ads is not None:
                datastore.ads = self.__fill_ads(datastore.ads, class_id_path_dict)

            if datastore.schema_nodes is not None:
                datastore.schema_nodes = self.__fill_schema_nodes(datastore.schema_nodes, class_id_path_dict)

            if datastore.schema_classes is not None:
                for schema_class in datastore.schema_classes:
                    class_path = class_id_path_dict.get(schema_class.class_id)
                    if class_path is None:
                        # There is no class or user does not have Read permissions
                        schema_class.class_name = "Permission denied"
                        schema_class.class_path = ["Permission denied"]
                    elif len(class_path) > 0:
                        schema_class.class_name = class_path[-1]
                        schema_class.class_path = class_path
                    else:
                        schema_class.class_path = class_path

        return datastore

    def __load_ads_with_class_info(self, schema: AbstractDataSchema) -> Union[DtlError, AbstractDataSchema]:
        schema_nodes = list(map(lambda node_def: node_def.node, schema.get_nodes()))
        class_id_path_dict = self.__load_paths_of_class_ids(schema_nodes)

        if isinstance(class_id_path_dict, DtlError):
            return class_id_path_dict

        return self.__fill_ads(schema, class_id_path_dict)

    def __fill_ads(self, schema: AbstractDataSchema, class_id_path_dict: Dict[UUID, List[str]]) -> Union[
        DtlError, AbstractDataSchema]:
        schema_nodes = list(map(lambda node_def: node_def.node, schema.get_nodes()))
        for schema_node in schema_nodes:
            classifications = schema_node.classifications
            for classification in classifications:
                class_path = class_id_path_dict.get(classification.class_id)
                if class_path is None:
                    # There is no class or user does not have Read permissions
                    classification.class_name = "Permission denied"
                    classification.class_path = ["Permission denied"]
                elif len(class_path) > 0:
                    classification.class_name = class_path[-1]
                    classification.class_path = class_path
                else:
                    classification.class_path = class_path

        return schema

    def __load_schema_nodes_with_class_info(self, schema_nodes: List[SchemaNode]) -> Union[DtlError, List[SchemaNode]]:
        class_id_path_dict = self.__load_paths_of_class_ids(schema_nodes)
        return self.__fill_schema_nodes(schema_nodes, class_id_path_dict)

    def __fill_schema_nodes(self, schema_nodes: List[SchemaNode], class_id_path_dict: Dict[UUID, List[str]]) -> Union[DtlError, List[SchemaNode]]:
        for schema_node in schema_nodes:
            classifications = schema_node.classifications
            for classification in classifications:
                class_path = class_id_path_dict.get(classification.class_id)
                if class_path is None:
                    # There is no class or user does not have Read permissions
                    classification.class_name = "Permission denied"
                    classification.class_path = ["Permission denied"]
                elif len(class_path) > 0:
                    classification.class_name = class_path[-1]
                    classification.class_path = class_path
                else:
                    classification.class_path = class_path
        return schema_nodes

    def __load_paths_of_class_ids(self, schema_nodes: List[SchemaNode]) -> Union[DtlError, Dict[str, List[str]]]:
        class_id_set = set([])
        for schema_node in schema_nodes:
            classifications = schema_node.classifications
            for classification in classifications:
                class_id_set.add(classification.class_id)

        res = []
        if len(class_id_set) > 0:
            res = self.http_client.make_authed_request(
                f'/yggy/entities?page=1&size={len(class_id_set)}',
                HttpMethod.PUT,
                list(map(lambda x: str(x), class_id_set)))

            if isinstance(res, DtlError):
                return res

        class_id_path_dict = {}
        for classInfo in res:
            class_id = UUID(classInfo["id"])
            class_path = classInfo["path"]
            class_id_path_dict[class_id] = class_path

        return class_id_path_dict

    def print(self, datastore_id: UUID, full_path: bool = False) -> Union[DtlError, None]:
        """
        TEMPORARILY DISABLED.
        From the provided id, get the corresponding Datastore schema

        :param datastore_id: id of the datastore to be retrieved
        :return:
        """
        print("Feature currently disabled, please contact Datalogue.")
        # res = self.http_client.make_authed_request(self.service_uri + "/datastores/" + str(datastore_id) + "/schema", HttpMethod.GET)
        # if isinstance(res, DtlError):
        #     if res.message.__contains__("Route not found"):
        #         res.message = "Invalid ID, please verify that your data store exists with dtl.datastore.list()"
        #     return res
        # schema = Schema._from_payload(res)
        # schema.print(full_path)

    def delete(self, datastore_id: UUID) -> Union[DtlError, bool]:
        """
        Deletes the given Datastore

        :param datastore_id: id of the datastore to be deleted
        :return: true if successful, false otherwise
        """

        res = self.http_client.make_authed_request(
            self.service_uri + "/datastores/" + str(datastore_id), HttpMethod.DELETE)

        if isinstance(res, DtlError):
            return res
        else:
            return True

    def download(self, datastore_id: UUID, file_format: FileFormat, path: str, limit: Optional[int] = None) -> Union[
        DtlError, str]:
        """
        Downloads the resource in a file

        :param datastore_id: id of the resource to download
        :param file_format: File format of the downloaded resource
        :param path: location to save the file
        :param limit: limit the number of elements to be downloaded
        :return: An error or the path to the downloaded file
        """

        params = {
            "fileFormat": file_format.value,
        }

        if limit is not None:
            if not isinstance(limit, int) or limit < 0:
                return DtlError("limit should be a positive integer")

            params["limit"] = limit

        res = self.http_client.make_authed_request(
            self.service_uri + "/datastores/" + str(datastore_id) + "/download",
            HttpMethod.GET, params=params, stream=True
        )

        if isinstance(res, DtlError):
            return res

        with open(path, 'wb') as f:
            for chunk in res.iter_content(chunk_size=1024 * 32):
                if chunk:  # filter out keep-alive new chunks
                    f.write(chunk)

        return path

    def load_arrow_table(self, datastore_id: UUID, limit: Optional[int] = None, batch_size: int = 1024) -> \
            Union[DtlError, Table]:
        """
        Loads the resource in an in memory arrow buffer.
        This can be used either with Nvidia RAPIDS or pandas

        pandas:
        df = dtl.client.datastore.load_arrow_table(datastore_id).to_pandas()

        :param datastore_id: if of the resource to load
        :param limit: elements to be loaded in the frame
        :param batch_size: size of the batch to be retrieved at a time from the collection
        :return:
        """

        params = {
            "fileFormat": FileFormat.Csv.value
        }

        if limit is not None:
            if not isinstance(limit, int) or limit < 0:
                return DtlError("limit should be a positive integer")

            params["limit"] = limit

        res = self.http_client.make_authed_request(
            self.service_uri + "/datastores/" + str(datastore_id) + "/download",
            HttpMethod.GET, params=params, stream=True
        )

        if isinstance(res, DtlError):
            return res

        stream = ResponseStream(res.iter_content(batch_size))

        return csv.read_csv(stream)

    def upload(self, datastore_id: UUID, file_format: FileFormat, path: str, params: dict = {}) -> Union[
        DtlError, None]:
        """
        Enable to upload a file to the specified resource through the platform

        :param datastore_id: id of the resource to upload the data through
        :param file_format: format of the file to be uploaded
        :param path: location of the file to upload on disk
        :return:
        """

        with open(path, 'rb') as f:
            files = [MultiPartFile("content", f)]

            data = [
                MultiPartData("file-format", file_format.value),
                MultiPartData("params", "{}"),
            ]

            res = self.http_client.post_multipart_data(
                self.service_uri + "/datastores/" + str(datastore_id) + "/upload",
                files, data
            )

            if isinstance(res, DtlError):
                return res

    def classify_fields(self, datastore_id: UUID, classifier_id: Union[str, UUID]) -> Union[DtlError, UUID]:
        """
        Classify the fields of a datastore. Classifications will be found within the ‘classifications’ property of
         each SchemaNode of the datastore.

        :param datastore_id: the id of the datastore to classify the fields of
        :param classifier: the classifier to apply classes to the data points, which will be used to calculate the
         classes of the fields
        :return: the job id of the pipeline classifying the datastore fields
        """

        req_params = {}
        req_params['classifier'] = str(classifier_id)
        res = self.http_client.make_authed_request(
            self.service_uri + "/datastores/" + str(datastore_id) + "/classify",
            HttpMethod.POST,
            params=req_params)

        if isinstance(res, DtlError):
            return res

        return UUID(res["jobId"])

    def add_classifications(self, datastore_id: UUID, classifications: List[NodeClassification]) -> Union[
        DtlError, List[SchemaNode]]:
        """
        Apply classes to specified fields (append, not override)

        :param datastore_id: the id of the datastore to add classes to
        :param classifications: a list of NodeClassifications, representing the fields to add classes to, and the classes to add
        :return: a list of altered SchemaNodes if successful, or DtlError if failed
        """

        res = self.http_client.make_authed_request(
            self.service_uri + "/datastores/" + str(datastore_id) + "/classifications/add",
            HttpMethod.PUT,
            list(map(lambda x: x._as_payload(), classifications)))

        if isinstance(res, DtlError):
            return res

        updated_schema_nodes = _parse_list(SchemaNode._schema_node_from_payload)(res)
        if isinstance(updated_schema_nodes, DtlError):
            return updated_schema_nodes

        return self.__load_schema_nodes_with_class_info(updated_schema_nodes)

    def remove_classifications(self, datastore_id: UUID, classifications: List[NodeClassification]) -> Union[
        DtlError, List[SchemaNode]]:
        """
        Remove classes from specified fields.

        :param datastore_id: the id of the datastore to remove classes from
        :param classifications: a list of NodeClassifications, representing the fields from which classes should be removed, and the classes  to remove from those fields
        :return: a list of altered SchemaNodes if successful, or DtlError if failed
        """

        res = self.http_client.make_authed_request(
            self.service_uri + "/datastores/" + str(datastore_id) + "/classifications/remove",
            HttpMethod.PUT,
            list(map(lambda x: x._as_payload(), classifications)))

        if isinstance(res, DtlError):
            return res

        updated_schema_nodes = _parse_list(SchemaNode._schema_node_from_payload)(res)
        if isinstance(updated_schema_nodes, DtlError):
            return updated_schema_nodes

        return self.__load_schema_nodes_with_class_info(updated_schema_nodes)

    def override_classifications(self, datastore_id: UUID, classifications: List[NodeClassification]) -> Union[
        DtlError, List[SchemaNode]]:
        """
        Remove all classes from specified fields, and then apply new classes to those fields.

        :param datastore_id: the id of the datastore whose fields should be overridden
        :param classifications: a list of NodeClassifications, representing the fields that should be cleared and the new classes for  those fields
        :return: a list of altered SchemaNodes if successful, or DtlError if failed
        """

        res = self.http_client.make_authed_request(
            self.service_uri + "/datastores/" + str(datastore_id) + "/classifications/replace",
            HttpMethod.PUT,
            list(map(lambda x: x._as_payload(), classifications)))

        if isinstance(res, DtlError):
            return res

        updated_schema_nodes = _parse_list(SchemaNode._schema_node_from_payload)(res)
        if isinstance(updated_schema_nodes, DtlError):
            return updated_schema_nodes

        return self.__load_schema_nodes_with_class_info(updated_schema_nodes)



from typing import Union, Optional
from uuid import UUID

from datalogue.clients.http import _HttpClient, HttpMethod
from datalogue.models.organization import User, _user_from_payload, _users_from_payload
from datalogue.errors import DtlError
from datalogue.dtl_utils import _parse_list


class _UserClient:
    """
    Client to interact with the User objects
    """

    def __init__(self, http_client: _HttpClient):
        self.http_client = http_client

    def get_current_user(self) -> Union[DtlError, User]:
        """
        Get information for current user

        :return: Error if it fails or user otherwise
        """
        res = self.http_client.make_authed_request("/user", HttpMethod.GET)

        if isinstance(res, DtlError):
            return DtlError("Could not retrieve user.", res)

        return _user_from_payload(res)

    def generate_reset_password_url(self, user_id: UUID) -> Union[DtlError, str]:
        """
        Get url which can be used to create new user password

        :return: Error if it fails or url otherwise
        """
        res = self.http_client.make_authed_request(f"/password/{user_id}/request-change", HttpMethod.POST)

        if isinstance(res, DtlError):
            return DtlError("Could not retrieve reset url.", res)
        return res

    def get(self, user_id: UUID) -> Union[DtlError, User]:
        """
        Get public user information: first_name, last_name, email
        for a given user_id .

        :param user_id: UUID
        :return: Error if it fails or user otherwise
        """

        params = {
            "id": str(user_id)
        }

        res = self.http_client.make_authed_request(f"/user/public", HttpMethod.GET, params=params)

        if isinstance(res, DtlError):
            return DtlError("Could not retreive user.", res)

        return _users_from_payload(res)

    def search_by_email(self, text_in_email: str, page: int = 1, item_per_page: int = 25) -> Union[DtlError, User]:
        """
        Search for info of the user (within the organization) given a text in the user email. The search will look for
        that string in the email. (i.e.: will include the user `some_email@domain.com` if you search for the text `some`)

        :param text_in_email:
        :param page: page to be retrieved (default 1)
        :param item_per_page: number of items to be put in a page (default 25)
        :return: The users containing that text in the email field (will be empty if there is no users with that text
        in the email) or an Error as a string
        """

        res = self.http_client.make_authed_request(
            f"/organization/users/search?email={text_in_email}&page={page}&size={item_per_page}",
            HttpMethod.GET
        )

        if isinstance(res, DtlError):
            return DtlError("Could not retreive users.", res)

        return _parse_list(_users_from_payload)(res)



from typing import Union

from datalogue.errors import DtlError, _enum_parse_error
from datalogue.dtl_utils import SerializableStringEnum
from datalogue.models.scope_level import Scope
from uuid import UUID


class Permission(SerializableStringEnum):
    """
    Class that handles all permission types
    """
    Share = "Share"
    Write = "Write"
    Read = "Read"

    @staticmethod
    def parse_error(s: str) -> str:
        return DtlError(_enum_parse_error("permission type", s))

    @staticmethod
    def from_string(string: str) -> Union[DtlError, 'Permission']:
        return SerializableStringEnum.from_str(Permission)(string)


# TODO change to the standard Permission Model (write, read, share)
class OntologyPermission(SerializableStringEnum):
    """
    Class that handles Ontology permission types
    """
    Write = "Write"
    Read = "Read"

    @staticmethod
    def parse_error(s: str) -> str:
        return DtlError(_enum_parse_error("Ontology permission type", s))

    @staticmethod
    def from_string(string: str) -> Union[DtlError, 'OntologyPermission']:
        return SerializableStringEnum.from_str(OntologyPermission)(string)


# TODO change to the standard Permission Model (write, read, share)
class CredentialPermission(SerializableStringEnum):
    """
    Class that handles Credential permission types
    """
    Write = "Write"
    Use = "Use"
    Read = "Read"

    @staticmethod
    def parse_error(s: str) -> str:
        return DtlError(_enum_parse_error("permission type", s))

    @staticmethod
    def from_string(string: str) -> Union[DtlError, 'CredentialPermission']:
        return SerializableStringEnum.from_str(CredentialPermission)(string)


class ObjectType(SerializableStringEnum):
    """
    Class that handles object types
    """
    Regex = "Regex"
    Classifier = "Classifier"
    Tag = "Tag"
    PipelineTemplate = "PipelineTemplate"

    @staticmethod
    def parse_error(s: str) -> str:
        return DtlError(_enum_parse_error("object type", s))

    @staticmethod
    def from_string(string: str) -> Union[DtlError, 'ObjectType']:
        return SerializableStringEnum.from_str(ObjectType)(string)


class SharePermission:
    """
    This is a class that represents an object permission.

    Attributes:
        target_id (UUID): the uuid of the User, Group or Organization..
        target_type (Scope): the target you are sharing with. It can be a User, Group or Organization.
        permission (Permission): the type of permission you want to grant. It can be Read, Write or Share.
    """

    def __init__(
            self,
            object_type: ObjectType,
            target_id: UUID = None,
            target_type: Scope = None,
            permission: Permission = None
    ):
        """
        The constructor for Share class.

        Parameters:
            object_type (ObjectType): the type of object
            target_id (UUID): the uuid of the User, Group or Organization..
            target_type (Scope): the target you are sharing with. It can be a User, Group or Organization.
            permission (Permission): the type of permission you want to grant. It can be Read, Write or Share.
        """

        self.object_type = object_type
        self.target_id = target_id
        self.target_type = target_type
        self.permission = permission

    def __repr__(self):
        return(f'{self.permission!r} permissions for this {self.object_type!r} are extended to '
               f'target {self.target_type!r} ' 
               f'with id: {self.target_id!r}')

    def from_payload(self, payload: dict) -> Union[DtlError, 'SharePermission']:

        target_id = payload.get("targetId")
        if target_id is None or not isinstance(target_id, str):
            return DtlError("'targetId' is missing or not a string")

        target_type = payload.get("targetType")
        if target_type is None or not isinstance(target_type, str):
            return DtlError("'targetType' is missing or not a string")

        permission = payload.get("permission")
        if permission is None or not isinstance(permission, str):
            return DtlError("'permission' is missing or not a string")

        return SharePermission(self.object_type, target_id, target_type, permission)

class UnsharePermission:
    """
    This is a class that represents an object permission removal.
    Attributes:
        target_id (UUID): the uuid of the User, Group or Organization..
        target_type (Scope): the target you are unsharing with. It can be a User, Group or Organization.
        permission (Permission): the type of permission you want to remove. It can be Read, Write or Share.
    """

    def __init__(
            self,
            object_type: ObjectType,
            target_id: UUID = None,
            target_type: Scope = None,
            permission: Permission = None
    ):
        """
        The constructor for Unshare class.
        Parameters:
            object_type (ObjectType): the type of object
            target_id (UUID): the uuid of the User, Group or Organization..
            target_type (Scope): the target you are unsharing with. It can be a User, Group or Organization.
            permission (Permission): the type of permission you want to remove. It can be Read, Write or Share.
        """

        self.object_type = object_type
        self.target_id = target_id
        self.target_type = target_type
        self.permission = permission

    def __repr__(self):
        return(f'{self.permission!r} permissions for this {self.object_type!r} are withdrawn from '
               f'target {self.target_type!r} '
               f'with id: {self.target_id!r}')

    def from_payload(self, payload: dict) -> Union[DtlError, 'UnsharePermission']:

        target_id = payload.get("targetId")
        if target_id is None or not isinstance(target_id, str):
            return DtlError("'targetId' is missing or not a string")

        target_type = payload.get("targetType")
        if target_type is None or not isinstance(target_type, str):
            return DtlError("'targetType' is missing or not a string")

        permission = payload.get("permission")
        if permission is None or not isinstance(permission, str):
            return DtlError("'permission' is missing or not a string")

        return UnsharePermission(self.object_type, target_id, target_type, permission)



from datalogue.clients.http import _HttpClient, Union, Optional, HttpMethod
from datalogue.clients.jobs import _JobsClient
from datalogue.clients.datastore_collections import _DatastoreCollectionClient
from datalogue.clients.credentials import _CredentialsClient
from datalogue.clients.datastore import _DatastoreClient
from datalogue.clients.organization import _OrganizationClient
from datalogue.clients.group import _GroupClient
from datalogue.clients.tag import _TagClient
from datalogue.clients.user import _UserClient
from datalogue.clients.ontology import _OntologyClient
from datalogue.clients.ml import _MLClient
from datalogue.clients.classifier import _ClassifierClient
from datalogue.clients.regex import _RegexClient
from datalogue.clients.pipeline import _PipelineClient
from datalogue.errors import DtlError

from urllib.parse import urlparse
import os

from datalogue.models.organization import User
from datalogue.models.version import Version


def _get_ssl_verify_env() -> bool:
    verify_certificate = os.environ.get('DTL_SSL_VERIFY_CERT')
    return verify_certificate is None or verify_certificate.lower() != 'false'


class DtlCredentials:
    """
    Information to be able to connect to the platform

    if verify_certificate is set, this variable will be used
    otherwise we look at the environment variable `DTL_SSL_VERIFY_CERT`
    if none is set then we use True as a default

    :param username: username to be used to login on the platform
    :param password: password to be used to login on the platform
    :param uri: root url where the system lives ie: https://test.datalogue.io/api
    :param verify_certificate: (Optional) verify the certificate when connecting to remote, no value means true.
    """

    def __init__(self, username: str, password: str, uri: str, verify_certificate: Optional[bool] = None):
        self.username = username.strip()
        self.password = password.strip()

        if verify_certificate is None:
            verify_certificate = _get_ssl_verify_env()

        self.verify_certificate = verify_certificate

        uri = uri.strip()

        if self.validate_url(uri) is not True:
            raise DtlError("The URL you provided is invalid")

        if not uri.endswith("/api"):
            raise DtlError("The URL you provided doesn't end with '/api' it is most likely invalid")

        self.uri = uri

    def validate_url(self, uri: str) -> bool:
        parsed_url = urlparse(uri)
        return all([parsed_url.scheme, parsed_url.netloc])

    def __repr__(self):
        res = f'{self.__class__.__name__}(username: {self.username!r}, password: ****, uri: {self.uri!r})'
        if self.verify_certificate is None:
            return res
        else:
            return res[:-1] + f', verify_certificate: {self.verify_certificate!r})'


class Dtl:
    """
    Root class to be built to interact with all the services

    :param credentials: contains the information to connect
    """

    def __init__(self, credentials: DtlCredentials):
        self.uri = credentials.uri
        self.username = credentials.username
        self.http_client = _HttpClient(credentials.uri, credentials.verify_certificate)

        login_res = self.http_client.login(credentials.username, credentials.password)
        if isinstance(login_res, DtlError):
            raise login_res

        self.group = _GroupClient(self.http_client)
        """Client to interact with the groups"""
        self.user = _UserClient(self.http_client)
        """Client to interact with the users"""
        self.organization = _OrganizationClient(self.http_client)
        """Client to interact with the organization part of the stack"""
        self.jobs = _JobsClient(self.http_client)
        """Client to interact with the jobs"""
        self.datastore_collection = _DatastoreCollectionClient(self.http_client)
        """Client to interact with the datastore collections"""
        self.datastore = _DatastoreClient(self.http_client)
        """Client to interact with the datastores"""
        self.credentials = _CredentialsClient(self.http_client)
        """Client to interact with credentials"""
        self.ontology = _OntologyClient(self.http_client)
        """Client to interact with ontologies"""
        self.ml = _MLClient(self.http_client)
        """Client to interact with ml"""
        self.classifier = _ClassifierClient(self.http_client, self.ontology)
        """Client to interact with classifiers"""
        self.regex = _RegexClient(self.http_client)
        """Client to interact with regexes"""
        self.tag = _TagClient(self.http_client)
        """Client to interact with tags"""
        self.pipeline = _PipelineClient(self.http_client)
        """Client to interact with the stream collections"""

    def __repr__(self):
        return f'Logged in {self.uri!r} with {self.username!r} account.'

    @staticmethod
    def signup(uri='', first_name='', last_name='', email='', password='', accept_terms=True,
               verify_certificate: Optional[bool] = None
               ) -> Union[DtlError, User]:
        """
        Perform signup of user
        :param uri: The target URI where the user data will be associated in
        :param accept_terms: Whether the user accept the following terms : https://www.datalogue.io/pages/terms-of-service
        :return: User object if successful, else return DtlError
        """

        if verify_certificate is None:
            verify_certificate = _get_ssl_verify_env()

        http_client = _HttpClient(uri, verify_certificate=verify_certificate)
        http_client.get_csrf()
        registered_user = http_client.signup(first_name, last_name, email, password, accept_terms)
        return registered_user

    def version(self) -> Union[DtlError, Version]:
        """
        Get version number of SDK, platform, and the platform's services'
        :return: Version object containing version number of SDK, platform, and the platform's services'
        """
        res = self.http_client.make_authed_request("/version", HttpMethod.GET)
        if isinstance(res, DtlError):
            return res
        return Version.from_payload(res)





from typing import Optional, Union
from datalogue.clients.http import _HttpClient, HttpMethod
from datalogue.anduin.models.stream import Stream, StreamStatus
from datalogue.errors import DtlError
from uuid import UUID


class _AnduinClient:
    """
    Client to interact with the Anduin APIs
    """

    def __init__(self, http_client: _HttpClient):
        self.http_client = http_client

    def run(self, stream: Stream, stream_id: Optional[UUID]=None) -> Union[DtlError, UUID]:
        """
        Run given stream directly through Anduin
        :param stream: Stream to run
        :param stream_id: Id to be used by the stream
        :return: Returns stream id
        """

        params = {}
        if id is not None:
            params = {"id": stream_id}

        res = self.http_client.make_authed_request('/run', HttpMethod.POST, stream._as_payload(), params=params)

        if isinstance(res, DtlError):
            return res

        return res["streamId"]

    def status(self, stream_id: UUID) -> Union[DtlError, StreamStatus]:
        """
        Returns the status of given stream id
        :param stream_id: Id of the stream
        :return: Returns status of the stream
        """
        res = self.http_client.make_authed_request('/status/' + str(stream_id), HttpMethod.GET)

        if isinstance(res, DtlError):
            return res

        return StreamStatus.from_payload(res)



from abc import ABC, abstractmethod
from typing import Dict, List, Set, Optional, Union
from uuid import UUID

from datalogue.errors import _enum_parse_error, DtlError
from datalogue.models.kinesis.shards import ShardAttributes
from datalogue.models.schema import AbstractDataSchema, SchemaNode, Classification
from datalogue.dtl_utils import _parse_list, _parse_string_list, SerializableStringEnum


class NodeClassification:
    """
    The mapping of a class to a field, as applied using the add, remove, and override commands of the dtl.datastore client.

    :param schema_node_id: the id of a schema field
    :param ontology_node_id: the class of a schema field
    :param score: the Confidence score for the classification, from 0.0 - 1.0 (defaults to 1.0)
    """

    def __init__(self, schema_node_id: Union[str, UUID], ontology_node_id: Union[str, UUID], score: float = 1.0):
        self.schema_node_id = schema_node_id
        self.ontology_node_id = ontology_node_id
        self.score = score

    def __repr__(self):
        return (f'{self.__class__.__name__}('
                f'schema_node_id: {self.schema_node_id!r}, '
                f'ontology_node_id: {self.ontology_node_id!r},' 
                f'score: {self.score!r}, ')

    def _as_payload(self) -> Union[DtlError, dict]:
        """
        Dictionary representation of the object with camelCase keys
        :return:
        """
        return {
            "schemaNodeId": str(self.schema_node_id),
            "classId": str(self.ontology_node_id),
            "score": str(self.score)
        }


class CredentialType(SerializableStringEnum):
    S3 = "S3"
    GCS = "GCS"
    Azure = "Azure"
    JDBC = "Jdbc"
    Mongo = "Mongo"
    Http = "Http"
    Socrata = "Socrata"
    FileSystem = "FileSystem"
    Void = "Void"
    Kinesis = "Kinesis"
    Hadoop = "Hadoop"
    Kafka = "Kafka"
    Minio = "Minio"
    HadoopAzure = "HadoopAzure"

    @staticmethod
    def parse_error(s: str) -> DtlError:
        return DtlError(_enum_parse_error("store type", s))

    @staticmethod
    def credential_type_from_str(string: str) -> Union[DtlError, 'CredentialType']:
        return SerializableStringEnum.from_str(CredentialType)(string)


class EncryptionType(SerializableStringEnum):
    AES256 = "AES256"
    KeyId = "KeyId"
    CustomerKey = "CustomerKey"

    @staticmethod
    def parse_error(s: str) -> DtlError:
        return DtlError(_enum_parse_error("encryption type", s))

    @staticmethod
    def encryption_type_from_str(string: str) -> Union[DtlError, 'EncryptionType']:
        return SerializableStringEnum.from_str(EncryptionType)(string)


class FileFormat(SerializableStringEnum):
    Json = "Json"
    Csv = "Csv"
    Text = "Text"
    Xml = "Xml"
    Html = "Html"
    Excel = "Excel"
    Avro = "Avro"
    Parquet = "Parquet"

    @staticmethod
    def parse_error(s: str) -> DtlError:
        return DtlError(_enum_parse_error("file format", s))

    @staticmethod
    def file_format_from_str(string: str) -> Union[DtlError, 'FileFormat']:
        return SerializableStringEnum.from_str(FileFormat)(string)


class DatastoreDef(ABC):
    type_field = "type"
    id_field = "id"

    def __init__(self, credential_type: CredentialType, datastore_id: Optional[UUID] = None):
        self.type = credential_type
        self.datastore_id = datastore_id
        super().__init__()

    def __eq__(self, other: 'DatastoreDef'):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def _base_payload(self) -> dict:
        base = [(DatastoreDef.type_field, self.type.value)]
        if self.datastore_id is not None:
            base.append((DatastoreDef.id_field, str(self.datastore_id)))
        return dict(base)

    @abstractmethod
    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object
        :return:
        """
        pass


class Cell:
    """
        Describes a cell in html table
    """

    def __init__(self, label: str, span: int, index: int, total: int, tag: str, tagOverrides:[] = None ):
        """
        Builds a cell

        :param label: text to display
        :param span: the html column span attribute
        :param index: index of this cell in the row
        :param total: total number of cells in the row
        """
        self.label = label
        self.span = span
        self.index = index
        self.total = total
        self.tag = tag
        self.tagOverrides = tagOverrides

    def __repr__(self):
        return f"{self.__class__.__name__}(label: {self.label!r}, span: {self.span!r}, \
                index: {self.index!r}, total: {self.total!r}, tag: {self.tag!r}, tagOverrides: {self.tagOverrides!r})"


def _cell_from_payload(json: dict) -> Union[DtlError, Cell]:
    label = json.get("label")
    if label is None:
        return DtlError("Cell has to have a 'label' key")

    span = json.get("span")
    if span is None:
        return DtlError("Cell has to have a 'span' key")

    index = json.get("index")
    if index is None:
        return DtlError("Cell has to have an 'index' key")

    total = json.get("total")
    if total is None:
        return DtlError("Cell has to have a 'total' key")

    return Cell(label, span, index, total, json.get("tag"), json.get("tagOverrides"))

class Datastore:
    """
    Represents a pointer to a datastore
    """

    def __init__(self,
                 name: str,
                 definition: DatastoreDef,
                 credential_id: Optional[UUID] = None,
                 alias: Optional[str] = None,
                 datastore_id: Optional[UUID] = None,
                 samples: Optional[List[List[Cell]]] = None,
                 schema_paths: Optional[List[List[str]]] = None,
                 schema_classes: Optional[List[List[str]]] = None,
                 schema_labels: Optional[Set[str]] = None,
                 schema_nodes: Optional[List[SchemaNode]] = None,
                 ads: Optional[AbstractDataSchema] = None):
        """
        Builds a pointer to a datastore

        :param name: name of the pointer
        :param alias: unique alias accross the organization for the pointer
        :param definition: definition on how to access the data
        :param credential_id: reference to the Credential to access the data
        :param datastore_id: defined if the object was persisted, otherwise None
        :param samples: list of samples that were found in the data source
        :param schema_paths: Used to replace paths options on the cli
        :param schema_classes: Used to replace classes options on the cli
        :param schema_labels: Set of labels that exists in schema
        """
        self.id = datastore_id
        self.name = name
        self.alias = alias
        self.definition = definition
        self.definition.datastore_id = datastore_id
        self.credential_id = credential_id
        self.samples = samples
        self.schema_paths = schema_paths
        self.schema_classes = schema_classes
        self.schema_labels = schema_labels
        self.schema_nodes = schema_nodes
        self.ads = ads

    def __eq__(self, other: 'Datastore'):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def __repr__(self):
        return f'{self.__class__.__name__}(id: {self.id}, name: {self.name!r}, alias: {self.alias!r}, ' \
               f'credential_id: {self.credential_id}, definition: {self.definition!r}, ' \
               f'samples: {self.samples!r}, schema_paths: {self.schema_paths!r}, schema_labels: {self.schema_labels!r}, schema_nodes: {self.schema_nodes!r})'

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object
        :return:
        """
        base = {
            "name": self.name,
            "gate": self.definition._as_payload(),
        }
        if self.id is not None:
            base["id"] = str(self.id)

        if self.alias is not None:
            base["alias"] = self.alias

        if self.credential_id is not None:
            base["credentialsId"] = str(self.credential_id)

        return base


###############################################################################
#                              Cloud Storage
###############################################################################


class S3DatastoreDef(DatastoreDef):
    type_str = CredentialType.S3

    def __init__(self,
                 bucket: str,
                 key: str,
                 file_format: FileFormat,
                 params: Optional[Dict[str, str]] = None,
                 datastore_id: Optional[UUID] = None):
        DatastoreDef.__init__(self, S3DatastoreDef.type_str, datastore_id)
        self.bucket = bucket
        self.key = key
        self.file_format = file_format
        self.params = params

    def __repr__(self):
        return f'{self.__class__.__name__}(bucket: {self.bucket!r}, key: {self.key!r}, ' \
               f'file_format: {self.file_format!r}, params: {self.params})'

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object with camelCase keys
        :return:
        """
        base = self._base_payload()
        base["bucket"] = self.bucket
        base["key"] = self.key
        base["format"] = self.file_format.value
        if self.params is not None:
            base["params"] = self.params
        return base


def _s3_datastore_def_from_payload(d: dict) -> Union[DtlError, S3DatastoreDef]:
    type_field = d.get(DatastoreDef.type_field)
    if not isinstance(type_field, str):
        return DtlError("string %s is missing from the json" % (
            DatastoreDef.type_field))

    if type_field != S3DatastoreDef.type_str.value:
        return DtlError("The object %s is not an S3 definition" % (str(d)))

    bucket = d.get("bucket")
    if bucket is None:
        return DtlError("'bucket' needs to be defined in an S3 definition")

    key = d.get("key")
    if key is None:
        return DtlError("'key' needs to be defined in an S3 definition")

    file_format = d.get("format")
    if file_format is None:
        return DtlError("'file_format' needs to be defined in an S3 definition")
    else:
        file_format = FileFormat.file_format_from_str(file_format)
        if isinstance(file_format, DtlError):
            return file_format

    params = d.get("params")
    datastore_id = d.get(DatastoreDef.id_field)

    return S3DatastoreDef(bucket, key, file_format, params, datastore_id)


class GCSDatastoreDef(DatastoreDef):
    type_str = CredentialType.GCS

    def __init__(self,
                 bucket: str,
                 file_name: str,
                 file_format: FileFormat,
                 params: Optional[Dict[str, str]] = None,
                 datastore_id: Optional[UUID] = None):
        DatastoreDef.__init__(self, GCSDatastoreDef.type_str, datastore_id)
        self.bucket = bucket
        self.file_name = file_name
        self.file_format = file_format
        self.params = params

    def __repr__(self):
        return f'{self.__class__.__name__}(bucket: {self.bucket!r}, file_name: {self.file_name!r}, ' \
               f'file_format: {self.file_format!r}, params: {self.params})'

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object  with camelCase keys
        :return:
        """
        base = self._base_payload()
        base["bucket"] = self.bucket
        base["fileName"] = self.file_name
        base["format"] = self.file_format.value
        if self.params is not None:
            base["params"] = self.params
        return base


def _gcs_datastore_def_from_payload(d: dict) -> Union[DtlError, GCSDatastoreDef]:
    type_field = d.get(DatastoreDef.type_field)
    if not isinstance(type_field, str):
        return DtlError("string %s is missing from the json" % (DatastoreDef.type_field))

    if type_field != GCSDatastoreDef.type_str.value:
        return DtlError("The object %s is not a GCS definition" % (str(d)))

    bucket = d.get("bucket")
    if bucket is None:
        return DtlError("'bucket' needs to be defined in a GCS definition")

    file_name = d.get("fileName")
    if file_name is None:
        return DtlError("'fileName' needs to be defined in a GCS definition")

    file_format = d.get("format")
    if file_format is None:
        return DtlError("'file_format' needs to be defined in a GCS definition")
    else:
        file_format = FileFormat.file_format_from_str(file_format)
        if isinstance(file_format, DtlError):
            return file_format

    params = d.get("params")
    store_id = d.get(DatastoreDef.id_field)

    return GCSDatastoreDef(bucket, file_name, file_format, params, store_id)


class AzureDatastoreDef(DatastoreDef):
    type_str = CredentialType.Azure

    def __init__(self,
                 container: str,
                 file_name: str,
                 file_format: FileFormat,
                 params: Optional[Dict[str, str]] = None,
                 datastore_id: Optional[UUID] = None):
        DatastoreDef.__init__(self, AzureDatastoreDef.type_str, datastore_id)
        self.container = container
        self.file_name = file_name
        self.file_format = file_format
        self.params = params

    def __repr__(self):
        return f'{self.__class__.__name__}(container: {self.container!r}, file_name: {self.file_name!r}, ' \
               f'file_format: {self.file_format!r}, params: {self.params})'

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object with camelCase keys
        :return:
        """
        base = self._base_payload()
        base["container"] = self.container
        base["fileName"] = self.file_name
        base["format"] = self.file_format.value
        if self.params is not None:
            base["params"] = self.params
        return base


def _azure_datastore_def_from_payload(d: dict) \
        -> Union[DtlError, AzureDatastoreDef]:
    type_field = d.get(DatastoreDef.type_field)
    if not isinstance(type_field, str):
        return DtlError("string %s is missing from the json" % DatastoreDef.type_field)

    if type_field != AzureDatastoreDef.type_str.value:
        return DtlError("The object %s is not an Azure Datastore definition" % (str(d)))

    container = d.get("container")
    if container is None:
        return DtlError("'container' needs to be defined in an Azure Datastore definition")

    file_name = d.get("fileName")
    if file_name is None:
        return DtlError("'fileName' needs to be defined in an Azure Datastore definition")

    file_format = d.get("format")
    if file_format is None:
        return DtlError("'file_format' needs to be defined in an Azure Datastore definition")
    else:
        file_format = FileFormat.file_format_from_str(file_format)
        if isinstance(file_format, DtlError):
            return file_format

    params = d.get("params")
    datastore_id = d.get(DatastoreDef.id_field)

    return AzureDatastoreDef(container, file_name, file_format, params, datastore_id)


class DatalogueDatastoreDef(DatastoreDef):
    type_str = CredentialType.Minio

    def __init__(self,
                 bucket: str,
                 key: str,
                 file_format: FileFormat,
                 params: Optional[Dict[str, str]] = None,
                 datastore_id: Optional[UUID] = None):
        DatastoreDef.__init__(self, DatalogueDatastoreDef.type_str, datastore_id)
        self.bucket = bucket
        self.key = key
        self.file_format = file_format
        self.params = params

    def __repr__(self):
            return f'{self.__class__.__name__}(bucket: {self.bucket!r}, key: {self.key!r}, ' \
                f'file_format: {self.file_format!r}, params: {self.params})'

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object with camelCase keys
        :return:
        """
        base = self._base_payload()
        base["bucket"] = self.bucket
        base["key"] = self.key
        base["format"] = self.file_format.value
        if self.params is not None:
            base["params"] = self.params
        return base

    @staticmethod
    def _from_payload(d: dict) -> Union[DtlError, 'DatalogueDatastoreDef']:
        type_field = d.get(DatastoreDef.type_field)
        if not isinstance(type_field, str):
            return DtlError("string %s is missing from the json" % (
                DatastoreDef.type_field))

        if type_field != DatalogueDatastoreDef.type_str.value:
            return DtlError("The object %s is not an Minio definition" % (str(d)))

        bucket = d.get("bucket")
        if bucket is None:
            return DtlError("'bucket' needs to be defined in an Minio definition")

        key = d.get("key")
        if key is None:
            return DtlError("'key' needs to be defined in an Minio definition")

        file_format = d.get("format")
        if file_format is None:
            return DtlError("'file_format' needs to be defined in an Minio definition")
        else:
            file_format = FileFormat.file_format_from_str(file_format)
            if isinstance(file_format, DtlError):
                return file_format

        params = d.get("params")
        datastore_id = d.get(DatastoreDef.id_field)

        return DatalogueDatastoreDef(bucket, key, file_format, params, datastore_id)


class HadoopDatastoreDef(DatastoreDef):
    type_str = CredentialType.Hadoop

    def __init__(self,
                 location: str,
                 file_format: FileFormat,
                 params: Optional[Dict[str, str]] = None,
                 datastore_id: Optional[UUID] = None):
        DatastoreDef.__init__(self, HadoopDatastoreDef.type_str, datastore_id)
        self.location = location
        self.file_format = file_format
        self.params = params

    def __repr__(self):
        return f'{self.__class__.__name__}(location: {self.location!r}, ' \
               f'file_format: {self.file_format!r}, params: {self.params})'

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object with camelCase keys
        :return:
        """
        base = self._base_payload()
        base["location"] = self.location
        base["format"] = self.file_format.value
        if self.params is not None:
            base["params"] = self.params
        return base

    @staticmethod
    def _from_payload(d: dict) -> Union[DtlError, 'HadoopDatastoreDef']:
        type_field = d.get(DatastoreDef.type_field)
        if not isinstance(type_field, str):
            return DtlError("string %s is missing from the json" % (
                DatastoreDef.type_field))

        if type_field != HadoopDatastoreDef.type_str.value:
            return DtlError("The object %s is not a Hadoop definition" % (str(d)))

        location = d.get("location")
        if location is None:
            return DtlError("'location' needs to be defined")

        file_format = d.get("format")
        if file_format is None:
            return DtlError("'format' needs to be defined")
        else:
            file_format = FileFormat.file_format_from_str(file_format)
            if isinstance(file_format, DtlError):
                return file_format
            if file_format != FileFormat.Parquet and file_format != FileFormat.Avro:
                return DtlError("'format' needs to be Avro or Parquet")

        params = d.get("params")
        resource_id = d.get(DatastoreDef.id_field)

        return HadoopDatastoreDef(location, file_format, params, resource_id)


class KafkaDatastoreDef(DatastoreDef):
    type_str = CredentialType.Kafka

    def __init__(self,
                 topic: str,
                 format: FileFormat,
                 params: Optional[Dict[str, str]] = None,
                 datastore_id: Optional[UUID] = None):
        DatastoreDef.__init__(self, KafkaDatastoreDef.type_str, datastore_id)
        self.topic = topic
        self.format = format
        self.params = params

    def __repr__(self):
        return f'{self.__class__.__name__}(topic: {self.topic!r}, ' \
            f'format: {self.format!r}, params: {self.params})'

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object with camelCase keys
        :return:
        """
        base = self._base_payload()
        base["topic"] = self.topic
        base["format"] = self.format.value
        if self.params is not None:
            base["params"] = self.params
        return base

    @staticmethod
    def _from_payload(d: dict) -> Union[DtlError, 'KafkaDatastoreDef']:
        type_field = d.get(KafkaDatastoreDef.type_field)
        if not isinstance(type_field, str):
            return DtlError("string %s is missing from the json" % (
                DatastoreDef.type_field))

        if type_field != KafkaDatastoreDef.type_str.value:
            return DtlError("The object %s is not a Kafka definition" % (str(d)))

        topic = d.get("topic")
        if topic is None:
            return DtlError("'topic' needs to be defined")

        format = d.get("format")
        if format is None:
            return DtlError("'format' needs to be defined")
        else:
            format = FileFormat.file_format_from_str(format)
            if isinstance(format, DtlError):
                return format
            if format not in (FileFormat.Csv, FileFormat.Text, FileFormat.Json):
                return DtlError("'format' needs to be CSV, TEXT, Json")

        params = d.get("params")
        resource_id = d.get(DatastoreDef.id_field)

        return KafkaDatastoreDef(topic, format, params, resource_id)


###############################################################################
#
#                              Database Connectors
###############################################################################

class JdbcDatastoreDef(DatastoreDef):
    type_str = CredentialType.JDBC

    def __init__(self,
                 root_table: str,
                 datastore_id: Optional[UUID] = None):
        DatastoreDef.__init__(self, JdbcDatastoreDef.type_str, datastore_id)
        self.root_table = root_table

    def _as_payload(self) -> dict:
        """Dictionary representation of the object with camelCase keys"""
        base = self._base_payload()
        base["rootTable"] = self.root_table
        return base


def _jdbc_datastore_def_from_payload(d: dict) -> Union[DtlError, JdbcDatastoreDef]:
    type_field = d.get(DatastoreDef.type_field)
    if not isinstance(type_field, str):
        return DtlError("string %s is missing from the json" % DatastoreDef.type_field)

    if type_field != JdbcDatastoreDef.type_str.value:
        return DtlError("The object %s is not a Jdbc definition" % str(d))

    root_table = d.get("rootTable")
    if root_table is None:
        return DtlError("'rootTable' needs to be defined in a Jdbc definition")

    datastore_id = d.get(DatastoreDef.id_field)

    return JdbcDatastoreDef(root_table, datastore_id)


class MongoDatastoreDef(DatastoreDef):
    type_str = CredentialType.Mongo

    def __init__(self,
                 collection: str,
                 datastore_id: Optional[UUID] = None):
        DatastoreDef.__init__(self, MongoDatastoreDef.type_str, datastore_id)
        self.collection = collection

    def _as_payload(self) -> dict:
        """Dictionary representation of the object with camelCase keys"""
        base = self._base_payload()
        base["collection"] = self.collection
        return base


def _mongo_datastore_def_from_payload(d: dict) -> Union[DtlError, MongoDatastoreDef]:
    type_field = d.get(DatastoreDef.type_field)
    if not isinstance(type_field, str):
        return DtlError("string %s is missing from the json" % DatastoreDef.type_field)

    if type_field != MongoDatastoreDef.type_str.value:
        return DtlError("The object %s is not a Mongo definition" % str(d))

    collection = d.get("collection")
    if collection is None:
        return DtlError("'collection' needs to be defined in a Mongo definition")

    datastore_id = d.get(DatastoreDef.id_field)

    return MongoDatastoreDef(collection, datastore_id)


class KinesisDatastoreDef(DatastoreDef):
    """
    Creates a datastore that references a Kinesis dataset
    """
    type_str =  CredentialType.Kinesis

    def __init__(self,
                 stream_name: str,
                 file_format: FileFormat,
                 shard_attributes: ShardAttributes = ShardAttributes(),
                 params: Optional[Dict[str, str]]=None,
                 datastore_id: Optional[UUID]=None):
        DatastoreDef.__init__(self, KinesisDatastoreDef.type_str, datastore_id)
        self.stream_name = stream_name
        self.file_format = file_format
        self.shard_attributes = shard_attributes
        self.params = params

    def __repr__(self):
        return f'{self.__class__.__name__}(stream_name: {self.stream_name!r}, file_format: {self.file_format!r}, ' \
               f'shard_attributes: {self.shard_attributes!r}, params: {self.params})'

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object with camelCase keys
        :return:
        """
        base = self._base_payload()
        base["streamName"] = self.stream_name
        base["format"] = self.file_format.value
        base['shardAttributes'] = self.shard_attributes._as_payload()
        if self.params is not None:
            base["params"] = self.params
        return base

    @staticmethod
    def _from_payload(d: dict) -> Union[DtlError, 'KinesisDatastoreDef']:
        type_field = d.get(DatastoreDef.type_field)
        if not isinstance(type_field, str):
            return DtlError("string %s is missing from the json" % (DatastoreDef.type_field))

        if type_field != KinesisDatastoreDef.type_str.value:
            return DtlError("The object %s is not an Kinesis definition" % (str(d)))

        stream_field = d.get("streamName")
        if stream_field is None:
            return DtlError("'streamName': needs to be defined")


        format_field = d.get("format")
        if format_field is None:
            return DtlError("'format': needs to be defined")
        else:
            format_field = FileFormat.file_format_from_str(format_field)
            if isinstance(format_field, DtlError):
                return format_field

        shard_attributes = d.get("shardAttributes")
        if shard_attributes is None:
            return DtlError("'shardAttributes': needs to be defined")
        else:
            shard_attributes = ShardAttributes._from_payload(shard_attributes)

        params = d.get("params")
        datastore_id = d.get(DatastoreDef.id_field)

        return KinesisDatastoreDef(stream_field, format_field, shard_attributes, params, datastore_id)

class KafkaDatastoreDef(DatastoreDef):
    type_str = CredentialType.Kafka

    def __init__(self,
                 topic: str,
                 format: FileFormat,
                 params: Optional[Dict[str, str]] = None,
                 datastore_id: Optional[UUID] = None):
        DatastoreDef.__init__(self, KafkaDatastoreDef.type_str, datastore_id)
        self.topic = topic
        self.format = format
        self.params = params

    def __repr__(self):
        return f'{self.__class__.__name__}(topic: {self.topic!r}, ' \
            f'format: {self.format!r}, params: {self.params})'

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object with camelCase keys
        :return:
        """
        base = self._base_payload()
        base["topic"] = self.topic
        base["format"] = self.format.value
        if self.params is not None:
            base["params"] = self.params
        return base

    @staticmethod
    def _from_payload(d: dict) -> Union[DtlError, 'KafkaDatastoreDef']:
        type_field = d.get(KafkaDatastoreDef.type_field)
        if not isinstance(type_field, str):
            return DtlError("string %s is missing from the json" % (
                DatastoreDef.type_field))

        if type_field != KafkaDatastoreDef.type_str.value:
            return DtlError("The object %s is not a Kafka definition" % (str(d)))

        topic = d.get("topic")
        if topic is None:
            return DtlError("'topic' needs to be defined")

        file_format = d.get("format")
        if file_format is None:
            return DtlError("'format' needs to be defined")
        else:
            file_format = FileFormat.file_format_from_str(file_format)
            if isinstance(file_format, DtlError):
                return file_format
            if file_format not in (FileFormat.Csv, FileFormat.Text, FileFormat.Json):
                return DtlError("'format' needs to be CSV, TEXT, Json")

        params = d.get("params")
        resource_id = d.get(DatastoreDef.id_field)

        return KafkaDatastoreDef(topic, file_format, params, resource_id)


###############################################################################
#                              API Sources
###############################################################################


class SocrataDatastoreDef(DatastoreDef):
    """
    Creates a datastore that references a Socrata dataset
    """
    type_str = CredentialType.Socrata

    def __init__(self,
                 socrata_id: str,
                 datastore_id: Optional[UUID] = None):
        """
        :param socrata_id: id of the dataset on socrata
        :param datastore_id: id of the datastore on the datalogue platform. This will only be set if the object was
         retrieved from the platform
        """
        DatastoreDef.__init__(self, SocrataDatastoreDef.type_str, datastore_id)
        self.socrata_id = socrata_id

    def _as_payload(self) -> dict:
        """Dictionary representation of the object with camelCase keys"""
        base = self._base_payload()
        base["socrataId"] = self.socrata_id
        return base

    @staticmethod
    def _from_payload(d: dict) -> Union[DtlError, 'SocrataDatastoreDef']:
        type_field = d.get(DatastoreDef.type_field)
        if not isinstance(type_field, str):
            return DtlError("string %s is missing from the json" % DatastoreDef.type_field)

        if type_field != SocrataDatastoreDef.type_str.value:
            return DtlError("The object %s is not a Socrata definition" % str(d))

        socrata_id = d.get("socrataId")
        if socrata_id is None:
            return DtlError("'socrata_id' needs to be defined in a Socrata definition")

        datastore_id = d.get(DatastoreDef.id_field)

        return SocrataDatastoreDef(socrata_id, datastore_id)


class HttpDatastoreDef(DatastoreDef):
    """
    Creates a datastore that references a Http dataset
    """
    type_str = CredentialType.Http

    def __init__(self,
                 url: str,
                 file_format: FileFormat,
                 headers: Optional[dict] = None,
                 query_params: Optional[dict] = None,
                 datastore_id: Optional[UUID] = None):
        """
        :param url: given url to fetch the content
        :param file_format: format of the file (json, csv, etc.)
        :param headers: headers to be added to the request (Accept, X-App-Token, etc.). it's optional
        :param query_params: http request parameters. it's optional
        """
        DatastoreDef.__init__(self, HttpDatastoreDef.type_str, datastore_id)
        self.url = url
        self.file_format = file_format
        self.headers = headers
        self.query_params = query_params

    def _as_payload(self) -> dict:
        """Dictionary representation of the object with camelCase keys"""
        base = self._base_payload()
        base["url"] = self.url
        base["format"] = self.file_format.value
        if self.headers is not None:
            base["headers"] = self.headers
        if self.query_params is not None:
             base["queryParams"] = self.query_params
        
        return base


def _http_datastore_def_from_payload(d: dict) -> Union[DtlError, HttpDatastoreDef]:
    type_field = d.get(DatastoreDef.type_field)
    if not isinstance(type_field, str):
        return DtlError("string %s is missing from the json" % DatastoreDef.type_field)

    if type_field != HttpDatastoreDef.type_str.value:
        return DtlError("The object %s is not an HttpResource definition" % str(d))

    url = d.get("url")
    if url is None:
        return DtlError("'url' needs to be defined in an HttpResource definition")

    file_format = d.get("format")
    if file_format is None:
        return DtlError("'file_format' needs to be defined in an HttpResource definition")
    else:
        file_format = FileFormat.file_format_from_str(file_format)
        if isinstance(file_format, DtlError):
            return file_format

    datastore_id = d.get(DatastoreDef.id_field)
    return HttpDatastoreDef(url, file_format, d.get("headers"), d.get("queryParams"), datastore_id)


###############################################################################
#                              File Sources
###############################################################################


class FileDatastoreDef(DatastoreDef):
    type_str = CredentialType.FileSystem

    def __init__(self,
                 location: str,
                 file_format: FileFormat,
                 params: Optional[Dict[str, str]] = None,
                 datastore_id: Optional[UUID] = None):
        DatastoreDef.__init__(self, FileDatastoreDef.type_str, datastore_id)
        self.location = location
        self.file_format = file_format
        self.params = params

    def _as_payload(self) -> dict:
        """Dictionary representation of the object with camelCase keys"""
        base = self._base_payload()
        base["location"] = self.location
        base["format"] = self.file_format.value
        if self.params is not None:
            base["params"] = self.params
        return base


def _file_datastore_def_from_payload(d: dict) -> Union[DtlError, FileDatastoreDef]:
    type_field = d.get(DatastoreDef.type_field)
    if not isinstance(type_field, str):
        return DtlError("string %s is missing from the json" % DatastoreDef.type_field)

    if type_field != FileDatastoreDef.type_str.value:
        return DtlError("The object %s is not a FileSystem definition" % str(d))

    location = d.get("location")
    if location is None:
        return DtlError("'location' needs to be defined in a FileSystem definition")

    file_format = d.get("format")
    if file_format is None:
        return DtlError("'file_format' needs to be defined in a FileSystem definition")
    else:
        file_format = FileFormat.file_format_from_str(file_format)
        if isinstance(file_format, DtlError):
            return file_format

    params = d.get("params")
    datastore_id = d.get(DatastoreDef.id_field)

    return FileDatastoreDef(location, file_format, params, datastore_id)


###############################################################################
#                              The Infinite Abyss
###############################################################################

class VoidDef(DatastoreDef):
    type_str = CredentialType.Void

    def __init__(self, datastore_id: Optional[UUID] = None):
        DatastoreDef.__init__(self, VoidDef.type_str, datastore_id)

    def __repr__(self):
        return "VoidDef()"

    def _as_payload(self) -> dict:
        """Dictionary representation of the object with camelCase keys"""
        return self._base_payload()


def _void_def_from_payload(d: dict) -> Union[DtlError, VoidDef]:
    type_field = d.get(DatastoreDef.type_field)
    if not isinstance(type_field, str):
        return DtlError("string %s is missing from the json" % DatastoreDef.type_field)

    if type_field != VoidDef.type_str.value:
        return DtlError("The object %s is not a Void definition" % str(d))

    datastore_id = d.get(DatastoreDef.id_field)

    return VoidDef(datastore_id)


_data_definitions = dict([
    (S3DatastoreDef.type_str.value, _s3_datastore_def_from_payload),
    (GCSDatastoreDef.type_str.value, _gcs_datastore_def_from_payload),
    (AzureDatastoreDef.type_str.value,_azure_datastore_def_from_payload),
    (JdbcDatastoreDef.type_str.value, _jdbc_datastore_def_from_payload),
    (MongoDatastoreDef.type_str.value, _mongo_datastore_def_from_payload),
    (HttpDatastoreDef.type_str.value, _http_datastore_def_from_payload),
    (SocrataDatastoreDef.type_str.value, SocrataDatastoreDef._from_payload),
    (FileDatastoreDef.type_str.value, _file_datastore_def_from_payload),
    (VoidDef.type_str.value, _void_def_from_payload),
    (KinesisDatastoreDef.type_str.value, KinesisDatastoreDef._from_payload),
    (KafkaDatastoreDef.type_str.value, KafkaDatastoreDef._from_payload),
    (HadoopDatastoreDef.type_str.value, HadoopDatastoreDef._from_payload),
    (DatalogueDatastoreDef.type_str.value, DatalogueDatastoreDef._from_payload)
])


def _datastore_def_from_payload(json: dict) -> Union[DtlError, DatastoreDef]:
    type_field = json.get(DatastoreDef.type_field)
    if type_field is None:
        return DtlError("The datastore definition object doesn't have a '%s' property" % \
                        DatastoreDef.type_field)

    parsing_function = _data_definitions.get(type_field)
    if parsing_function is None:
        return DtlError("Looks like '%s' datastore is not handled by the SDK" % \
               type_field)

    return parsing_function(json)


def _datastore_from_payload(json: dict) -> Union[DtlError, Datastore]:
    name = json.get("name")
    if name is None:
        return DtlError("'name' for a datastore should be defined")

    definition = json.get("definition")
    gate = json.get("gate")
    
    if definition is None and gate is None:
        return DtlError("Neither 'definition' nor 'gate' for a datastore is defined!'")
    elif definition is not None:
        definition = _datastore_def_from_payload(definition)
    else:
        definition = _datastore_def_from_payload(gate)

    alias = json.get("alias")
    credential_id = json.get("credentialsId")
    if credential_id is not None:
        credential_id = UUID(credential_id)
    datastore_id = json.get("id")

    samples = json.get("samples")
    if samples is not None:
        samples = _parse_list(_parse_list(_cell_from_payload))(samples)
        if isinstance(samples, DtlError):
            return samples

    schema_paths = json.get("schemaPaths")
    if schema_paths is not None:
        schema_paths = _parse_list(_parse_string_list)(schema_paths)
        if isinstance(schema_paths, DtlError):
            return schema_paths

    schema_classes = []
    classifications_json = json.get("classifications")
    if classifications_json is not None:
        classifications = _parse_list(Classification._from_payload)(classifications_json)
        if isinstance(classifications, DtlError):
            return classifications
        schema_classes = classifications

    schema_labels = None
    if schema_paths is not None:
        schema_labels = set()
        for row in schema_paths:
            for col in row:
                schema_labels.add(col)
        schema_labels = list(schema_labels)

    ads_json = json.get("ads")
    schema_nodes = None
    ads = None
    if ads_json is not None:
        schema_nodes = ads_json.get("nodes")
        if schema_nodes is not None:
            schema_nodes = _parse_list(SchemaNode._schema_node_from_payload)(schema_nodes)
            if isinstance(schema_nodes, DtlError):
                return schema_nodes
        ads = AbstractDataSchema._from_payload(ads_json, AbstractDataSchema.node_decoder)
    return Datastore(name, definition, credential_id, alias, datastore_id, samples, schema_paths, schema_classes, schema_labels, schema_nodes, ads)



from typing import List, Union, Optional
from datalogue.clients.http import _HttpClient, HttpMethod
from datalogue.models.datastore_collection import DatastoreCollection, _datastore_collection_from_dict
from datalogue.errors import DtlError
from uuid import UUID
from datalogue.dtl_utils import _parse_list
from functools import reduce


class _DatastoreCollectionClient:
    """
    Client to interact with the datastore collection objects
    """

    def __init__(self, http_client: _HttpClient):
        self.http_client = http_client
        self.service_uri = "/scout"

    def create(self, datastore_collection: DatastoreCollection) -> Union[DtlError, DatastoreCollection]:
        """
        Creates the Datastore Collection as specified.

        :param datastore_collection: Datastore Collection to be created
        :return: string with error message if failed, uuid otherwise
        """
        assert isinstance(datastore_collection, DatastoreCollection), 'Input is not a Datastore Collection instance.'
        # todo?? assert datastore indeed is a datastore

        res = self.http_client.make_authed_request(
            self.service_uri + "/datastore-collections", HttpMethod.POST,
            datastore_collection._as_payload())

        if isinstance(res, DtlError):
            return res

        return _datastore_collection_from_dict(res)

    def update(self, id: UUID, datastore_collection: DatastoreCollection) -> Union[DtlError, bool]:
        """
        Updates the Datastore Collection for the given id

        :param id: Id of the datastore collection to update
        :param datastore_collection: Datastore Collection to be updated
        :return: string with error message if failed, uuid otherwise
        """
        assert isinstance(datastore_collection, DatastoreCollection), 'Input is not a Datastore Collection instance.'

        res = self.http_client.make_authed_request(
            self.service_uri + "/datastore-collections/" + str(id), HttpMethod.PUT,
            datastore_collection._as_payload())


        if isinstance(res, DtlError):
            return res
        else:
            return True

    def list(self, page: int = 1, item_per_page: int = 25) -> Union[DtlError, List[DatastoreCollection]]:
        """
        List the Datastore Collection

        :param page: page to be retrieved
        :param item_per_page: number of datastore_collections to be put in a page
        :return: Returns a List of all the available datastore collections or an error message as a string
        """

        res = self.http_client.make_authed_request(
            self.service_uri + f"/datastore-collections?page={page}&size={item_per_page}", HttpMethod.GET, )

        if isinstance(res, DtlError):
            return res

        return _parse_list(_datastore_collection_from_dict)(res)

    def search(self,
               by_name: Optional[List[str]] = None,
               by_description: Optional[List[str]] = None,
               page: int = 1,
               item_per_page: int = 25) -> Union[DtlError, List[DatastoreCollection]]:
        """
        Search existing datastore collections for a given names. If nothing is defined for names, 
        it returns every datastore collections (basically matches all)

        :param by_name: a list of datastore names
        :param by_description: a list of datastore descriptions
        :param page: page to be retrieved
        :param item_per_page: number of items to be put in a page
        :return: Returns a List of all the available datastores or an error message as a string
        """
        queries = {}

        if isinstance(by_name, list):
            queries["name"] = by_name

        if isinstance(by_description, list):
            queries["description"] = by_description

        payload = {
            "queries": queries,
            "page": page,
            "size": item_per_page,
            "type": "datastore-collection"
        }
        
        res = self.http_client.make_authed_request(self.service_uri + "/v2/search", HttpMethod.POST, payload)

        if isinstance(res, DtlError):
            return res

        return _parse_list(_datastore_collection_from_dict)(res['results'])

    def get(self, id: UUID) -> Union[DtlError, DatastoreCollection]:
        """
        From the provided id, get the corresponding Datastore Collection

        :param id: id of the datastore collection to be fetched
        :return:
        """
        res = self.http_client.make_authed_request(
            self.service_uri + "/datastore-collections/" + str(id),
            HttpMethod.GET)

        if isinstance(res, DtlError):
            return res

        return _datastore_collection_from_dict(res)

    def delete(self, id: UUID) -> Union[DtlError, bool]:
        """
        Deletes the given Datastore Collection

        :param id: id of the datastore collection to be deleted
        :return: true if successful, the error otherwise
        """
        res = self.http_client.make_authed_request(
            self.service_uri + "/datastore-collections/" + str(id),
            HttpMethod.DELETE
        )

        if isinstance(res, DtlError):
            return res
        else:
            return True

    def add_datastore(
                    self,
                    datastore_collection_id: Union[str, UUID],
                    added_datastore_ids: List[Union[UUID, str]]
                        ) -> Union[DtlError, DatastoreCollection]:
        """
        Add datastores to a datastore collection. If one or more specified datastores already exists in the collection, this operation will ignore them.

        :param datastore_collection_id: id of the datastore collection that will be updated with new datastore(s)
        :param added_datastore_ids: a list of ids corresponding to the datastores to be added to the datastore collection
        :return: the updated DatastoreCollection object if successful, or DtlError if failed
        """

        res = self.http_client.make_authed_request(
            self.service_uri + "/datastore-collections/" + str(datastore_collection_id) + "/add-datastores",
            HttpMethod.PUT, added_datastore_ids
        )

        if isinstance(res, DtlError):
            return res

        return _datastore_collection_from_dict(res)

    def remove_datastore(
                    self,
                    datastore_collection_id: Union[str, UUID],
                    removed_datastore_ids: List[Union[UUID, str]]
                        ) -> Union[DtlError, DatastoreCollection]:
        """
        Remove datastores from a datastore collection. If one or more specified datastores is not present in the collection, this operation will ignore them.

        :param datastore_collection_id: id of the datastore collection that will be updated
        :param removed_datastore_ids: a list of ids corresponding to the datastores to be removed from the datastore collection
        :return: the updated DatastoreCollection object if successful, or DtlError if failed
        """

        res = self.http_client.make_authed_request(
            self.service_uri + "/datastore-collections/" + str(datastore_collection_id) + "/remove-datastores",
            HttpMethod.PUT, removed_datastore_ids
        )

        if isinstance(res, DtlError):
            return res

        return _datastore_collection_from_dict(res)



from typing import Optional


class DtlError(Exception):

    def __init__(self, message: str, cause: Optional['DtlError'] = None):
        # Call the base class constructor with the parameters it needs
        super(DtlError, self).__init__(message)
        self.message = message
        self.cause = cause

    def __repr__(self):
        return f"DtlError({self.message} {str(self.cause)})"

    def __eq__(self, other: 'DtlError'):
        if isinstance(self, other.__class__):
            return self.message == other.message
        return False


def _enum_parse_error(enum_description: str, string: str) -> str:
    return "'%s' is not a valid %s" % (string, enum_description)


def _property_not_found(property_name: str, obj: dict) -> DtlError:
    return DtlError(f"Could not find '{property_name}' in the dictionary: {obj!r}")


def _invalid_property_type(property_name: str, expected: str, obj: dict) -> DtlError:
    return DtlError(f"Property '{property_name}' was expected to be '{expected}' in {obj!r}")


def _invalid_parameter_error(reason: str, class_or_function: str) -> DtlError:
    return DtlError(f"Failed to call/create {class_or_function} because {reason}")


def _invalid_pagination_params(param_name: str) -> DtlError:
    return DtlError(f"The parameter '{param_name}' must be a positive integer")



import sys
from typing import Dict, Optional, Union, List
from uuid import UUID

import pandas as pd

from datalogue.errors import DtlError
from datalogue.models.graph import Graph, Node, NodeDef, Path
from datalogue.dtl_utils import _parse_string_list, _parse_list, _parse_datetime, _parse_uuid, map_option, SerializableStringEnum
from datetime import datetime
from datalogue.errors import _enum_parse_error

class ClassificationType(SerializableStringEnum):
    Automatic = "Automatic"
    Manual = "Manual"

    @staticmethod
    def parse_error(s: str) -> DtlError:
        return DtlError(_enum_parse_error("stream state", s))

    @staticmethod
    def classification_type_from_str(string: str) -> Union[DtlError, 'ClassificationType']:
        return SerializableStringEnum.from_str(ClassificationType)(string)


class Classification:
    """
    Represents the mapping and application of a class to a field. To edit classifications, please use the add, remove, and override commands of the dtl.datastore client. 
    """
    def __init__(self, class_id: UUID, classification_type: ClassificationType, class_name: str, class_path: List[str], ontology_id: UUID, classifier_id: Optional[UUID], score: float, created_at: datetime, created_by: Optional[UUID]):
        self.class_id = class_id
        self.classification_type = classification_type
        self.class_name = class_name
        self.class_path = class_path
        self.ontology_id = ontology_id
        self.classifier_id = classifier_id
        self.score = score
        self.created_at = created_at
        self.created_by = created_by

    def __repr__(self):
        return f"({self.class_name}, {self.score})"
    
    @staticmethod
    def _from_payload(payload: dict) -> Union[DtlError, 'Classification']:
        class_id = map_option(payload.get("classId"), _parse_uuid) 
        if class_id is None:
            return DtlError("Classification has to have a 'classId' key")

        if class_id is DtlError:
            return class_id

        classification_type = map_option(payload.get("classificationType"), ClassificationType.classification_type_from_str)
        if classification_type is None:
            return DtlError("Classification has to have a 'classificationType' key")

        if classification_type is DtlError:
            return classification_type

        ontology_id = map_option(payload.get("ontologyId"), _parse_uuid)
        if ontology_id is None:
            return DtlError("SchemaNode has to have a 'ontologyId' key")

        classifier_id = map_option(payload.get("classifierId"), _parse_uuid)

        score = payload.get("score")
        if score is None:
            return DtlError("Classification has to have a 'score' key")

        created_at = map_option(payload.get("createdAt"), _parse_datetime)
        if isinstance(created_at, DtlError):
            return created_at

        created_by = map_option(payload.get("createdBy"), _parse_uuid)
        if isinstance(created_by, DtlError):
            return created_by

        return Classification(class_id, classification_type, None, None, ontology_id, classifier_id, score, created_at, created_by)


class SchemaNode(Node):
    """
    Describes schema node
    """

    def __init__(self, id: UUID, label: str, data_type: str, classifications: List[Classification]):
        """
        Builds SchemaNode object

        :param id: schema node id
        :param label: column name
        :param data_type: type of the data
        """

        self.id = id
        self.label = label
        self.dataType = data_type
        self.classifications = classifications
        Node.__init__(self, id, label)

    def __repr__(self):
        classificationStr = ", ".join(list(map(lambda c: str(c), self.classifications)))
        return ('SchemaNode('
                    f'id: {self.id}, '
                    f'label: {self.label}, '
                    f'dataType: {self.dataType}, '
                    f'\n\tclassifications: [{classificationStr}]\n)')

    @staticmethod
    def _as_payload(self) -> Union[DtlError, dict]:
        """
        Dictionary representation of the object with camelCase keys
        :return:
        """

        return {
            "id": self.id,
            "label": self.label,
            "dataType": self.dataType
        }

    def _schema_node_from_payload(json: dict) -> Union[DtlError, 'SchemaNode']:
        node_id = json.get("id")
        if node_id is None:
            return DtlError("SchemaNode has to have a 'id' key")

        label = json.get("label")
        if label is None:
            return DtlError("SchemaNode has to have a 'label' key")

        dataType = json.get("dataType")
        classifications = _parse_list(Classification._from_payload)(json.get("classifications"))
        if isinstance(classifications, DtlError):
            return classifications

        return SchemaNode(node_id, label, dataType, classifications)

class GraphSampleNode(Node):
    """
    Describes schema node
    """

    def __init__(self, id: UUID, label: str, value: str):
        """
        Builds SchemaNode object

        :param id: schema node id
        :param label: column name
        :param value: value of the data
        """

        self.id = id
        self.label = label
        self.value = value
        Node.__init__(self, id, label)

    def _sample_node_from_payload(json: dict) -> Union[DtlError, 'GraphSampleNode']:
        node_id = json.get("id")
        if node_id is None:
            return DtlError("SampleNode has to have a 'id' key")

        label = json.get("label")
        if label is None:
            return DtlError("SampleNode has to have a 'label' key")

        value = json.get("value")
        if "value" not in json:
            return DtlError("SampleNode has to have a 'value' key")

        return GraphSampleNode(str(node_id), label, value)



class AbstractDataGraph(Graph):
    node_decoder = GraphSampleNode._sample_node_from_payload


class AbstractDataSchema(Graph):
    node_decoder = SchemaNode._schema_node_from_payload



class SchemaOntologyRow:
    """
    Represents printable representation of schema, its associated ontology & samples data.
    """

    def __init__(self, schema_node_def: NodeDef, samples: List[AbstractDataGraph]):
        self.schema_node_def = schema_node_def
        self.samples = samples

    def to_array(self, is_full_path: bool):
        def get_sample_val(index: int, full_path: Path):
            if len(self.samples) > index:
                sample = self.samples[index]
                return sample.path_to_node[full_path].node.value
            else:
                return None

        schema_node = self.schema_node_def.node
        full_path = self.schema_node_def.full_path
        sample_1_val = get_sample_val(0, full_path)
        sample_2_val = get_sample_val(1, full_path)
        sample_3_val = get_sample_val(2, full_path)
        ontology_node_id = schema_node.ontologyNodeId
        if ontology_node_id is None:
            ontology_node_id = "\"\""
        if not is_full_path:
            if schema_node.ontologyNodePath is not None and len(schema_node.ontologyNodePath) > 0:
                ontology_node_label = schema_node.ontologyNodePath[-1]
            else:
                ontology_node_label = "\"\""
            return [
                schema_node.id, 
                schema_node.label, 
                schema_node.dataType,
                ontology_node_label, 
                ontology_node_id, 
                sample_1_val, 
                sample_2_val,
                sample_3_val
            ]
        else:
            if schema_node.ontologyNodePath is not None:
                ontology_node_label = schema_node.ontologyNodePath
            else:
                ontology_node_label = ["\"\"", "\"\""]
            return [
                schema_node.id, 
                self.schema_node_def.full_path, 
                schema_node.dataType,
                ontology_node_label, 
                ontology_node_id, 
                sample_1_val, 
                sample_2_val,
                sample_3_val
            ]


class Schema:
    """
    Represents a schema.
    """

    def __init__(self, data_schema: AbstractDataSchema, samples: Optional[List[AbstractDataGraph]]):
        """
        Represents a schema structure

        :param data_schema: Represents data schema
        :param samples: A collection of samples data with the same structure as data_schema
        """

        self.data_schema = data_schema
        self.samples = samples

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, 'Schema']:
        schema = json.get("schema")
        samples_json = json.get("samples")
        if schema is None:
            return DtlError("Schema has to have a 'schema' key")
        if samples_json is None:
            return DtlError("Schema has to have a 'samples' key")
        ads = AbstractDataSchema._from_payload(schema, AbstractDataSchema.node_decoder)

        if samples_json is None:
            adg = None
        else:
            adg = []
            for obj in samples_json:
                parsed_obj = AbstractDataGraph._from_payload(obj, AbstractDataGraph.node_decoder)

                if isinstance(parsed_obj, DtlError):
                    return parsed_obj
                else:
                    adg.append(parsed_obj)
        return Schema(ads, adg)

    def print(self, full_path: bool):
        pd.set_option('max_colwidth', 29)
        nodes_def = self.data_schema.get_nodes()
        res = []
        if len(nodes_def) > 1000:
            return DtlError("Extremely large schema, please contact Datalogue support to enable larger schemas")
        for node_def in nodes_def:
            res.append(SchemaOntologyRow(node_def, self.samples).to_array(full_path))
        df = pd.DataFrame(pd.np.array(res),
                     columns=['schema_node_id', 'schema_node_name', 'schema_node_type', 'ontology_node_name',
                              'ontology_node_id', 'sample1', 'sample2', 'sample3'])
        sys.stdout.write(df.to_string(index=False))



from typing import List, Union, Optional
from datalogue.errors import DtlError, _property_not_found
from datalogue.models.transformations.commons import Transformation
from datalogue.dtl_utils import _parse_string_list


class ToInt(Transformation):
    """
      Allows to casts a node value to int
      """
    type_str = "ToInt"

    def __init__(self, path: List[str]):
        """
        Builds a Casting transformation to Int

        :param path: path to node to be cast to Int
        """
        Transformation.__init__(self, ToInt.type_str)
        self.path = path

    def __eq__(self, other: 'ToInt'):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def __repr__(self):
        return f"ToInt({self.path})"

    def _as_payload(self) -> dict:
        base = self._base_payload()
        base["path"] = self.path
        return base

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, 'ToInt']:
        path = json.get("path")
        if path is None:
            return _property_not_found("path", json)

        path = _parse_string_list(path)
        if isinstance(path, DtlError):
            return path

        return ToInt(path)


class ToDate(Transformation):
    """
    Allows to casts a node value to a date time with timezone.

    You can use the following symbols in the formatting string::

        Symbol  Meaning                     Presentation      Examples
        ------  -------                     ------------      -------
        G       era                         text              AD; Anno Domini; A
        u       year                        year              2004; 04
        y       year-of-era                 year              2004; 04
        D       day-of-year                 number            189
        M/L     month-of-year               number/text       7; 07; Jul; July; J
        d       day-of-month                number            10

        Q/q     quarter-of-year             number/text       3; 03; Q3; 3rd quarter
        Y       week-based-year             year              1996; 96
        w       week-of-week-based-year     number            27
        W       week-of-month               number            4
        E       day-of-week                 text              Tue; Tuesday; T
        e/c     localized day-of-week       number/text       2; 02; Tue; Tuesday; T
        F       week-of-month               number            3

        a       am-pm-of-day                text              PM
        h       clock-hour-of-am-pm (1-12)  number            12
        K       hour-of-am-pm (0-11)        number            0
        k       clock-hour-of-am-pm (1-24)  number            0

        H       hour-of-day (0-23)          number            0
        m       minute-of-hour              number            30
        s       second-of-minute            number            55
        S       fraction-of-second          fraction          978
        A       milli-of-day                number            1234
        n       nano-of-second              number            987654321
        N       nano-of-day                 number            1234000000

        V       time-zone ID                zone-id           America/Los_Angeles; Z; -08:30
        z       time-zone name              zone-name         Pacific Standard Time; PST
        O       localized zone-offset       offset-O          GMT+8; GMT+08:00; UTC-08:00;
        X       zone-offset 'Z' for zero    offset-X          Z; -08; -0830; -08:30; -083015; -08:30:15;
        x       zone-offset                 offset-x          +0000; -08; -0830; -08:30; -083015; -08:30:15;
        Z       zone-offset                 offset-Z          +0000; -0800; -08:00;
        julian  JDE Julian Date Format      text              120060 (equivalent to 29/02/2020, midnight time)
      """

    type_str = "ToDate"

    def __init__(self, path: List[str], date_format: Optional[List[str]] = None, optional_input: bool = False):
        """
        Builds a Casting transformation to a datetime with timezone

        :param path: path to node to be cast to Int
        :param date_format: List of formats to be used to attempt to parse the string. Defaults to [YYYY-MM-DDTHH:mm:ss.VVZ]
        """
        Transformation.__init__(self, ToDate.type_str)
        self.path = path
        self.date_format = date_format
        self.optional_input = optional_input

    def __eq__(self, other: 'ToDate'):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def __repr__(self):
        return f"ToDate({self.path!r}, {self.date_format!r}, {self.optional_input!r})"

    def _as_payload(self) -> dict:
        base = self._base_payload()
        base["path"] = self.path
        base["dateFormats"] = self.date_format
        base["optionalInput"] = self.optional_input
        return base

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, 'ToDate']:
        path = json.get("path")
        dateformats = json.get("dateFormats")
        if path is None:
            return _property_not_found("path", json)
        
        if dateformats is None:
            return _property_not_found("dateFormats", json)

        optional_input = json.get("optionalInput")
        if optional_input is None:
            optional_input = False

        path = _parse_string_list(path)
        dateformats = _parse_string_list(dateformats)
        if isinstance(path, DtlError):
            return path
        if isinstance(dateformats, DtlError):
            return dateformats

        return ToDate(path, dateformats, optional_input)



from typing import Dict, List, Optional, Union
from uuid import UUID
import itertools

from datalogue.errors import _enum_parse_error, DtlError
from datalogue.models.permission import OntologyPermission
from datalogue.models.scope_level import Scope
from datalogue.dtl_utils import _parse_list, SerializableStringEnum


class TrainingDataColumn:
    def __init__(self,
                 datastore_name: str,
                 path: List[str] = [],
                 count: int = 0):

        if count is not 0 and not isinstance(count, int):
            raise DtlError("count should be an integer in TrainingDataColumn")

        self.datastore_name = datastore_name
        self.path = path
        self.count = count

    def __eq__(self, other: 'TrainingDataColumn'):
        if isinstance(self, other.__class__):
            return (self.datastore_name == other.datastore_name and
                    self.path == other.path and
                    self.count == other.count)
        return False

    @staticmethod
    def _as_payload(training_data_column: 'TrainingDataColumn') -> Union[DtlError, dict]:
        payload = {
            "datastore_name": str(training_data_column.datastore_name),
            "path": list(map(lambda n: str(n), training_data_column.path)),
            "count": training_data_column.count
        }

        return payload

    @staticmethod
    def _from_payload(payload: dict) -> Union[DtlError, 'TrainingDataColumn']:
        datastore_name = payload.get("datastore_name")
        path = payload.get("path")
        count = payload.get("count")
        
        if datastore_name is None:
            return DtlError("'datastore_name' not defined in TrainingDataColumn")
        
        if path is None:
            path = []

        return TrainingDataColumn(datastore_name, path, count)


class OntologyNode:
    def __init__(self,
                 name: str,
                 description: Optional[str] = None,
                 children: List['OntologyNode'] = [],
                 id: Optional[UUID] = None,
                 total_training_count: int = 0,
                 training_data_columns: List[TrainingDataColumn] = []):

        if not isinstance(name, str):
            raise DtlError("name should be string in OntologyNode")

        if description is not None and not isinstance(description, str):
            raise DtlError("description should be string in OntologyNode")

        if not isinstance(children, List):
            raise DtlError("children should be list in OntologyNode")

        if id is not None and not isinstance(id, UUID):
            raise DtlError("id should be uuid in OntologyNode")

        if total_training_count is not 0 and not isinstance(total_training_count, int):
            raise DtlError("total_training_count should be an integer in OntologyNode")

        self.name = name
        self.description = description
        self.children = children
        self.id = id
        self.total_training_count = total_training_count
        self.training_data_columns = training_data_columns

    def __eq__(self, other: 'OntologyNode'):
        if isinstance(self, other.__class__):
            return (self.name == other.name and
                    self.description == other.description and
                    self.children == other.children and
                    self.id == other.id and 
                    self.total_training_count == other.total_training_count and 
                    self.training_data_columns == other.training_data_columns)
        return False


    def __repr__(self):
        return f'{self.__class__.__name__}(id: {self.id}, name: {self.name!r})'

    @staticmethod
    def as_payload(ontology_node: 'OntologyNode') -> Union[DtlError, dict]:
        payload = {
            "name": ontology_node.name,
            "children": list(map(lambda n: OntologyNode.as_payload(n), ontology_node.children)),
            "training_data_columns": list(map(lambda n: TrainingDataColumn._as_payload(n), ontology_node.training_data_columns)),
            "total_training_count": ontology_node.total_training_count
        }

        if ontology_node.id is not None:
            payload["node_id"] = str(ontology_node.id)

        if ontology_node.description is not None:
            payload["description"] = ontology_node.description

        return payload

    @staticmethod
    def from_payload(payload: dict) -> Union[DtlError, 'OntologyNode']:
        id = UUID(payload.get("id"))
        name = payload.get("name")
        description = payload.get("description")
        children = payload.get("children")
        total_training_count = payload.get("total_training_count")
        training_data_columns = payload.get("training_data_columns")

        if children is None:
            children = []
        else:
            children = _parse_list(OntologyNode.from_payload)(children)
        
        if training_data_columns is None:
            training_data_columns = []
        else:
            training_data_columns = _parse_list(OntologyNode.from_payload)(training_data_columns)
        
        if total_training_count is None:
            total_training_count = 0
        return OntologyNode(name, description, children, id, total_training_count, training_data_columns)


class Ontology:
    def __init__(self,
                 name: str,
                 description: Optional[str] = None,
                 tree: List[OntologyNode] = [],
                 id: Optional[UUID] = None):

        if not isinstance(name, str):
            raise DtlError("name should be string in Ontology")

        if description is not None and not isinstance(description, str):
            raise DtlError("description should be string in Ontology")

        if not isinstance(tree, List):
            raise DtlError("tree should be list in Ontology")

        if id is not None and not isinstance(id, UUID):
            raise DtlError("id should be uuid in Ontology")

        self.name = name
        self.description = description
        self.tree = tree
        self.id = id



    def __eq__(self, other: 'Ontology'):
        if isinstance(self, other.__class__):
            return (self.name == other.name and
                    self.description == other.description and
                    self.tree == other.tree and
                    self.id == other.id)
        return False

    def __repr__(self):
        def print_nodes(tree, output=[], level=0):
            for n in tree:
                padding = '   ' * level
                if level == 0:
                    output.append(padding + n.name)
                else:
                    output.append(padding + '|___' + n.name)
                for c in n.children:
                    print_nodes([c], output, level=level+1)
            return output

        first_line = f'Ontology(id: {self.id}, name: {self.name!r}, description: {self.description!r})' + '\n'
        return '\n'.join(print_nodes(self.tree, [first_line]))

    def leaves(self) -> List[OntologyNode]:
        def iterate(node: OntologyNode) -> List[OntologyNode]:
            if not node.children:
                return [ node ]
            else:
                return list(itertools.chain(*map(lambda n: iterate(n), node.children)))

        return list(itertools.chain(*map(lambda n: iterate(n), self.tree)))

    @staticmethod
    def _as_payload(ontology: 'Ontology') -> Union[DtlError, dict]:
        payload = {
            "name": ontology.name,
            "children": list(map(lambda n: OntologyNode.as_payload(n), ontology.tree))
        }

        if ontology.description is not None:
            payload["description"] = ontology.description

        return payload

    @staticmethod
    def _from_payload(payload: dict) -> Union[DtlError, 'Ontology']:
        id = UUID(payload.get("id"))
        name = payload.get("name")
        description = payload.get("description")
        tree = payload.get("tree")
        if tree is None:
            tree = []
        else:
            tree = _parse_list(OntologyNode.from_payload)(payload["tree"])
        return Ontology(name, description, tree, id=id)

    @staticmethod
    def _create_body_for_sharing(target_id: UUID, target_type: Scope, permission: OntologyPermission) -> Dict:

        if not isinstance(target_type, Scope):
             return DtlError(f"'{target_type}' should be a type of '{Scope.__name__}' and must be in: {list(map(str, Scope))}")

        if not isinstance(permission, OntologyPermission):
            return DtlError(f"'{permission}' should be a type of '{OntologyPermission.__name__}' and must be in: {list(map(str, OntologyPermission))}")

        body = {}

        body["organizations"] = None
        body["groups"] = None
        body["users"] = None

        if target_type == Scope.User:
            body["users"] = {str(target_id): permission.value}
        elif target_type == Scope.Group:
            body["groups"] = {str(target_id): permission.value}
        elif target_type == Scope.Organization:
            body["organizations"] = {str(target_id): permission.value}
        return body


    @staticmethod
    def _create_body_for_unsharing(target_id: UUID, target_type: Scope, permission: str) -> Dict:

        if not isinstance(target_type, Scope):
             return DtlError(f"'{target_type}' should be a type of '{Scope.__name__}' and must be in: {list(map(str, Scope))}")

        body = {}

        body["organizations"] = None
        body["groups"] = None
        body["users"] = None

        if target_type == Scope.User:
            body["users"] = {str(target_id): permission}
        elif target_type == Scope.Group:
            body["groups"] = {str(target_id): permission}
        elif target_type == Scope.Organization:
            body["organizations"] = {str(target_id): permission}
        return body


class DataRef:
    def __init__(self, node_id: Union[str, UUID], path_list: List[List[str]]):
        self.node_id = node_id
        self.path_list = path_list



from datetime import datetime, timedelta
from typing import List, Union
from uuid import UUID

from datalogue.clients.http import _HttpClient, HttpMethod
from datalogue.models.organization import Organization, _organization_from_payload, User, _users_from_payload, Group, \
    _group_from_payload, Domain, _domain_from_payload, CreateUserRequest
from datalogue.errors import DtlError
from datalogue.dtl_utils import _parse_list


class _OrganizationClient:
    """
    Client to interact with the Organization objects
    """
    errorMsg = "you may not have permission to do the action or the organization was not found"

    def __init__(self, http_client: _HttpClient):
        self.http_client = http_client

    def create(self, name: str) -> Union[DtlError, Organization]:
        """
        Creates an organization

        :param name: Organization to be created
        :return: string with error message if failed, the organization otherwise
        """
        res = self.http_client.make_authed_request(
            "/organization", HttpMethod.POST, {
                "name": name
            })

        if isinstance(res, DtlError):
            return res

        return _organization_from_payload(res)

    def get(self, org_id: UUID) -> Union[DtlError, Organization]:
        """
        From the provided id, get the corresponding Organization

        :param org_id: id of the organization to be retrieved
        :return: Error if it fails or Organization object otherwise
        """

        res = self.http_client.make_authed_request(
            f"/organization/{str(org_id)}/public", HttpMethod.GET)

        if isinstance(res, DtlError):
            return res

        return _organization_from_payload(res)

    def add_users(self, org_id: UUID, users: List[CreateUserRequest], expiry_date: datetime) -> Union[DtlError, List[str]]:
        """
        Add users to given organization and return reset password links for each user
        :param org_id: id of the organization
        :param users: list of CreateUserRequests
        :param expiry_date: expiration date of reset uri
        :return:
            List of set password link
            Error if expiration date was in past
            Error if there is email duplication or if at least one users with given email exists
        """
        body = {
            "users": [user.as_payload() for user in users],
            "expirationDate": expiry_date.strftime("%Y-%m-%dT%H:%M:%SZ")
        }
        res = self.http_client.make_authed_request(f"/organization/{str(org_id)}/users", HttpMethod.POST, body)
        return res

    def get_all_users(self, org_id: UUID, page: int = 1, item_per_page: int = 25) -> Union[DtlError, List[User]]:
        """
        From the provided organization id, get all users in the Organization

        :param org_id: id of the organization
        :param page: page to be retrieved
        :param item_per_page: number of users to be put in a page
        :return: Error if it fails or list of all users in the organization otherwise
        """

        res = self.http_client.make_authed_request(
            f"/organization/{str(org_id)}/users?page={page}&size={item_per_page}", HttpMethod.GET)
        if isinstance(res, DtlError):
            return DtlError(f"Could not retreive all users of the organization: {res.message}", _OrganizationClient.errorMsg)

        return _parse_list(_users_from_payload)(res)

    def get_groups(self, page: int = 1, item_per_page: int = 25) -> Union[DtlError, List[Group]]:
        """
        Get all groups in the current organization in which the user is a member

        :param page: page to be retrieved
        :param item_per_page: number of groups to be put in a page
        :return: Error if it fails or list of all groups in the organization otherwise
        """

        res = self.http_client.make_authed_request(
            f"/groups?page={page}&size={item_per_page}", HttpMethod.GET)

        if isinstance(res, DtlError):
            return DtlError(f"Could not retreive your groups of your organization: {res.message}")

        return _parse_list(_group_from_payload)(res)

    def get_all_groups(self, org_id: UUID, page: int = 1, item_per_page: int = 25) -> Union[DtlError, List[Group]]:
        """
        From the provided organization id, get all groups in the Organization when admin, or user groups for non admin

        :param org_id: id of the organization
        :param page: page to be retrieved
        :param item_per_page: number of groups to be put in a page
        :return: Error if it fails or list of groups
        """

        res = self.http_client.make_authed_request(
            f"/organization/{str(org_id)}/groups?page={page}&size={item_per_page}", HttpMethod.GET)

        if isinstance(res, DtlError):
            return DtlError(f"Could not retrieve all groups of the organization: {res.message}", _OrganizationClient.errorMsg)

        return _parse_list(_group_from_payload)(res)

    def add_domain(self, org_id: UUID, domain: str) -> Union[DtlError, bool]:
        """
        Add domain to provided organization

        :param org_id: id of the organization
        :return: Error if it fails or true if it's added
        """

        res = self.http_client.make_authed_request(
            f"/organization/{str(org_id)}/domain", HttpMethod.POST, {
                "domain": domain
            })

        if isinstance(res, DtlError):
            return res
        else:
            return True

    def remove_domain(self, org_id: UUID, domain: str) -> Union[DtlError, bool]:
        """
        Remove domain from provided organization

        :param org_id: id of the organization
        :return: Error if it fails or true if it's deleted
        """

        res = self.http_client.make_authed_request(
            f"/organization/{str(org_id)}/domain", HttpMethod.DELETE, {
                "domain": domain
            })

        if isinstance(res, DtlError):
            return res
        else:
            return True

    def get_all_domains(self, org_id: UUID, page: int = 1, item_per_page: int = 25) -> Union[DtlError, List[Domain]]:
        """
        Get all domains of the provided organization

        :param org_id: id of the organization
        :param page: page to be retrieved
        :param item_per_page: number of domains to be put in a page
        :return: Error if it fails or list of domains
        """

        res = self.http_client.make_authed_request(
            f"/organization/{str(org_id)}/domains?page={page}&size={item_per_page}", HttpMethod.GET)

        if isinstance(res, DtlError):
            return DtlError(f"Could not retreive all domains: {res.message}", _OrganizationClient.errorMsg)

        return _parse_list(_domain_from_payload)(res)




from typing import List, Union
from uuid import UUID

from datalogue.clients.http import _HttpClient, HttpMethod
from datalogue.models.organization import User, _users_from_payload, Group, _group_from_payload, Domain, _domain_from_payload
from datalogue.errors import DtlError
from datalogue.dtl_utils import _parse_list


class _GroupClient:
    """
    Client to interact with the Group objects
    """

    def __init__(self, http_client: _HttpClient):
        self.http_client = http_client

    def create(self, name: str, org_id: UUID) -> Union[DtlError, Group]:
        """
        Creates a group

        :param name: Group to be created
        :param org_id: Organization in which the group will be created
        :return: string with error message if failed, the group otherwise
        """
        res = self.http_client.make_authed_request(
            "/group", HttpMethod.POST, {
                "name": name,
                "organizationId": str(org_id)
            })

        if isinstance(res, DtlError):
            return DtlError("Could not create the group:", res)

        return _group_from_payload(res)

    def delete(self, group_id: UUID) -> Union[DtlError, bool]:
        """
        Delete group

        :param group_id: id of the group
        :return: Error if it fails or true if it's deleted
        """

        res = self.http_client.make_authed_request(
            f"/group/{str(group_id)}", HttpMethod.DELETE)

        if isinstance(res, DtlError):
            return DtlError("Could not delete the group", res)
        else:
            return True

    def get_users(self, group_id: UUID, page: int = 1, item_per_page: int = 25) -> Union[DtlError, List[User]]:
        """
        From the provided group id, get all users in the group

        :param group_id: id of the group
        :param page: page to be retrieved
        :param item_per_page: number of users to be put in a page
        :return: Error if it fails or list of all users in the group otherwise
        """

        res = self.http_client.make_authed_request(
            f"/group/{str(group_id)}/users?page={page}&size={item_per_page}", HttpMethod.GET)

        if isinstance(res, DtlError):
            return DtlError(f"Could not retreive all users of the group: {res.message}")

        return _parse_list(_users_from_payload)(res)

    def get_list(self, page: int = 1, item_per_page: int = 25) -> Union[DtlError, List[Group]]:
        """
        Get only a list of groups in which the user is a member (the whole database)
        :param page: page to be retrieved
        :param item_per_page: number of groups to be put in a page
        :return: Error if it fails or list of groups otherwise
        """

        res = self.http_client.make_authed_request(
            f"/groups?page={page}&size={item_per_page}", HttpMethod.GET)

        if isinstance(res, DtlError):
            return DtlError(f"Could not retreive all groups of the organization: {res.message}")

        return _parse_list(_group_from_payload)(res)

    def add_user(self, user_id: UUID, group_id: UUID) -> Union[DtlError, bool]:
        """
        Add provided user to provided group

        :param group_id: id of the group
        :param user_id: id of the user
        :return: Error if it fails or true if it's added
        """

        res = self.http_client.make_authed_request(
            f"/group/{str(group_id)}/user/{str(user_id)}", HttpMethod.POST)

        if isinstance(res, DtlError):
            return DtlError("Could not add user to the group:", res)
        else:
            return True

    def remove_user(self, user_id: UUID, group_id: UUID) -> Union[DtlError, bool]:
        """
        Remove user from provided group

        :param group_id: id of the group 
        :param user_id: id of the user
        :return: Error if it fails or true if it's deleted
        """

        res = self.http_client.make_authed_request(
            f"/group/{str(group_id)}/user/{str(user_id)}", HttpMethod.DELETE)

        if isinstance(res, DtlError):
            return DtlError("Could not delete user:", res)
        else:
            return True

    def add_right(self, group_id: UUID) -> Union[DtlError, bool]:
        """
        Add right to provided group : should have the right ManageOrganization

        :param group_id: id of the group
        :return: Error if it fails or true if it's added
        """

        res = self.http_client.make_authed_request(
            f"/group/{str(group_id)}/permission?name=ManageOrganization", HttpMethod.POST)

        if isinstance(res, DtlError):
            return DtlError("You may not have permission to do the action. ", res)
        else:
            return True

    def remove_right(self, group_id: UUID) -> Union[DtlError, bool]:
        """
        Remove right from provided group: should have the right ManageOrganization

        :param group_id: id of the group
        :return: Error if it fails or true if it's deleted
        """

        res = self.http_client.make_authed_request(
            f"/group/{str(group_id)}/permission?name=ManageOrganization", HttpMethod.DELETE)

        if isinstance(res, DtlError):
            return DtlError("You may not have permission to do the action.", res)
        else:
            return True



from datetime import datetime
from typing import Optional, Union, List
from uuid import UUID

from datalogue.errors import DtlError
from datalogue.dtl_utils import map_option, _parse_uuid, _parse_datetime


class Tag:
    """
    A unique tag of metadata that can be applied to datastores for targeted collection, viewing, and action orchestrations
    """

    def __init__(
            self,
            name: str,
            id: Optional[UUID] = None,
            created_at: Optional[datetime] = None,
            created_by: Optional[UUID] = None):
        self.id = id
        self.name = name
        self.created_at = created_at
        self.created_by = created_by

    def __eq__(self, other):
        return (self.name == other.name) and \
               (self.created_at == other.created_at) and \
               (self.created_by == other.created_by)

    def __repr__(self):
        return (f'{self.__class__.__name__}(\n'
                f'id: {self.id!r},\n '
                f'name: {self.name!r},\n'
                f'created_at: {self.created_at!r},\n'
                f'created_by: UUID({self.created_by!r})\n'
                f')')

    @staticmethod
    def _from_payload(payload: dict) -> Union[DtlError, 'Tag']:
        name = payload.get("name")
        if name is None:
            return DtlError("'name' is missing for a tag")

        id = map_option(payload.get("id"), _parse_uuid)
        created_at = payload.get("createdAt")
        created_by = payload.get("createdBy")
        if isinstance(created_at, DtlError):
            return created_at

        return Tag(name, id, created_at, created_by)



from typing import List, Union, Dict, Optional
from datalogue.errors import DtlError, _property_not_found, _enum_parse_error, _invalid_property_type
from datalogue.models.transformations.commons import Transformation, DataType
from datalogue.dtl_utils import _parse_string_list, SerializableStringEnum, _parse_list


class PickStrategy(SerializableStringEnum):
    """
    Data Types that can be specified in the pipeline
    """
    HighScore = "HighScore"
    Non = "None"
    Random = "Random"
    All = "All"

    @staticmethod
    def parse_error(s: str) -> str:
        return _enum_parse_error("PickStrategy", s)

    @staticmethod
    def from_str(string: str) -> Union[DtlError, 'PickStrategy']:
        return SerializableStringEnum.from_str(PickStrategy)(string)


class ClassNodeDescription:
    """
    Description of a node that will be selected from a class and that specifies its path location in the output.
    """
    def __init__(self, path: List[str], tag: str, pick_strategy: PickStrategy, data_type: DataType = DataType.String):
        """
        Builds a ClassNodeDescription to be used in the Structure transformation.

        :param path: path of the node to be created
        :param tag: class of the value to be picked up
        :param pick_strategy: How to pick up the value to be used
        :param data_type: Type of the output value
        """
        self.path = path
        self.tag = tag
        self.pick_strategy = pick_strategy
        self.data_type = data_type

    def __eq__(self, other: 'ClassNodeDescription'):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def __repr__(self):
        return f"ClassNodeDescription(path: {self.path}, tag: {self.tag}, strategy: {self.pick_strategy}, dataType: {self.data_type})"

    def _as_payload(self) -> dict:
        return {
            "type": "Class",
            "outputDataType": self.data_type.value,
            "outputPath": self.path,
            "class": self.tag,
            "pickStrategy": self.pick_strategy.value
        }

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, 'ClassNodeDescription']:
        if json.get("type") is None and json.get("type") != "Class":
            return DtlError("Invalid json payload type for ClassNodeDescription")

        path = json.get("outputPath")
        if path is None:
            return _property_not_found("outputPath", json)

        path = _parse_string_list(path)
        if isinstance(path, DtlError):
            return path

        tag = json.get("class")
        if tag is None:
            return _property_not_found("class", json)

        pick = json.get("pickStrategy")
        if pick is None:
            return _property_not_found("pickStrategy", json)

        pick = PickStrategy.from_str(pick)
        if isinstance(pick, DtlError):
            return pick

        data_type = json.get("outputDataType")
        if data_type is not None:
            data_type = DataType.from_str(data_type)
            if isinstance(data_type, DtlError):
                return data_type
        else:
            data_type = DataType.String

        return ClassNodeDescription(path, tag, pick, data_type)


class PathNodeDescription:
    """
    Instruction to copy a node from input into the output
    """
    def __init__(self, path: List[str], output_path: List[str], output_data_type: DataType = DataType.String,
                 is_output_array: bool = False, optional_input: bool = False):
        """
        Builds a PathNodeDescription to be used in the Structure transformation.

        :param path: path of the node to be used as input
        :param output_path: path of the node to be created in the output
        :param output_data_type: type for the node to be used as output
        :param is_output_array: will the output contain several nodes with identical paths
        :param optional_input:
            Sets if the input is mandatory for the structure. The default is to check
            the input data for the specified field and return and error if it is not present.
            if set to true, we will bypass that check and leave the path property untouched.
        """
        self.path = path
        self.output_path = output_path
        self.output_dataType = output_data_type
        self.is_output_array = is_output_array
        self.optional_input = optional_input

    def __eq__(self, other: 'PathNodeDescription'):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def __repr__(self):
        return f"PathNodeDescription(path={self.path!r}, output_path={self.output_path!r}, " \
               f"output_dataType={self.output_dataType!r}, is_output_array={self.is_output_array!r}," \
               f"optional_input={self.optional_input!r})"

    def _as_payload(self) -> dict:
        return {
            "type": "Path",
            "path": self.path,
            "outputPath": self.output_path,
            "outputDataType": self.output_dataType.value,
            "outputIsArray": self.is_output_array,
            "optionalInput": self.optional_input
        }

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, 'PathNodeDescription']:
        if json.get("type") is None and json.get("type") != "Path":
            return DtlError("Invalid json payload type for PathNodeDescription")

        path = json.get("path")
        if path is None:
            return _property_not_found("path", json)

        path = _parse_string_list(path)
        if isinstance(path, DtlError):
            return path

        output_path = json.get("outputPath")
        if output_path is None:
            return _property_not_found("outputPath", json)

        output_path = _parse_string_list(output_path)
        if isinstance(output_path, DtlError):
            return output_path

        is_output_array = json.get("is_output_array")
        if is_output_array is None:
            is_output_array = False

        output_data_type = json.get("outputDataType")
        if output_data_type is None:
            return _property_not_found("output_dataType", json)

        output_data_type = DataType.from_str(output_data_type)
        if isinstance(output_data_type, DtlError):
            return output_data_type

        optional_input = json.get("optionalInput")
        if optional_input is None:
            optional_input = False

        return PathNodeDescription(path, output_path, output_data_type, is_output_array, optional_input)


class Structure(Transformation):
    """
    Enables to pick nodes based on their class or path and organize them as you wish for the output

    When using a class to select a node, there are different strategies that can be used to pick it.

       * HighScore: Uses the node classified with the highest score
       * None: If several matches do not use any
       * Random: Just picks one
       * All: Picks all

    If no nodes with the specified class or path are found, the output node will be created with an empty value.

    If a cast fails, the element is dropped from the stream.
    """

    type_str = "Structure"

    def __init__(self, structure: List[Union[ClassNodeDescription, PathNodeDescription]]):
        """
        Builds StructureByClass transformation

        :param structure: List of nodes to include in the output
        """
        Transformation.__init__(self, Structure.type_str)
        self.structure = structure

    def __eq__(self, other: 'Structure'):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def __repr__(self):
        return f"Structure(structure: {self.structure!r})"

    def _as_payload(self) -> dict:
        base = self._base_payload()
        base["outputSchema"] = list(map(lambda s: s._as_payload(), self.structure))
        return base

    @staticmethod
    def _parse_class_or_path_description(json: Dict) -> Union[ClassNodeDescription, PathNodeDescription, DtlError]:
        description_type = json.get("type")
        if json.get("type") is None:
            return DtlError("Invalid node description inside the Structure transformation")

        if description_type == "Class":
            return ClassNodeDescription._from_payload(json)
        elif description_type == "Path":
            return PathNodeDescription._from_payload(json)
        else:
            return DtlError("Invalid type for the node description")

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, 'Structure']:
        structure = json.get("outputSchema")
        if structure is None:
            return _property_not_found("outputSchema", json)

        structure = _parse_list(Structure._parse_class_or_path_description)(structure)
        if isinstance(structure, DtlError):
            return structure

        return Structure(structure)



import abc
from base64 import b64decode
from typing import Union, List, Optional

from datalogue.errors import DtlError, _enum_parse_error
from datalogue.models.transformations.commons import ByClass, ByFieldName, Transformation
from datalogue.dtl_utils import _parse_list, SerializableStringEnum, _parse_string_list, EncryptionAlgorithm


class ObfuscationType(abc.ABC):
    """
    Abstract class used to reference the type of Obfuscate to apply to target data

    Contains the three subclasses, each representing an obfuscation type:
    - Encrypt, for two-way encryption
    - Hash, for one-way hash
    - Replace, for one-way string replacement
    """
    type_field = "type"

    def __init__(self, obfuscation_type: str):
        self.type = obfuscation_type
        super().__init__()

    def _base_payload(self) -> dict:
        return dict([(ObfuscationType.type_field, self.type)])

    @abc.abstractmethod
    def _as_payload(self) -> dict:
        """
        Represents the transformation in its dictionary construction
        :return: a dictionary representing this class
        """

    @abc.abstractmethod
    def _from_payload(json: dict) -> Union[DtlError, 'ObfuscationType']:
        """"""

class Obfuscate(Transformation):
    type_str = "Obfuscate"

    def __init__(
            self,
            targets: List[Union[ByClass, ByFieldName]],
            obfuscation_type: ObfuscationType
    ):
        """
        Obfuscate data with hashing, encryption, or string replacement.

        :param target: a list of references to data to target with the obfuscation
        :param obfuscation_type: the type of obfuscation to apply to the data specified by the target. Hash, encryption, or string replacement are available, each with additional parameters.
        """
        Transformation.__init__(self, Obfuscate.type_str)
        self.targets = targets
        self.obfuscation_type = obfuscation_type

    def __repr__(self):
        return ('Obfuscate('
                f'targets: {self.targets})'
                f'obfuscation_type: {self.obfuscation_type})')

    def _as_payload(self) -> dict:
        base = self._base_payload()
        targetClassesIds = []
        targetPaths = []

        for elem in self.targets:
            if isinstance(elem, ByClass):
                targetClassesIds.append(str(elem.class_id))
            elif isinstance(elem, ByFieldName):
                targetPaths.append(elem.field_name)

        base['targetClassesIds'] = targetClassesIds
        base['targetPaths'] = targetPaths
        base['obfuscationType'] = self.obfuscation_type._as_payload()
        return base

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, 'Obfuscate']:
        obfuscationType = ObfuscationType._from_payload(json.get("obfuscationType"))
        target_paths = _parse_list(_parse_string_list)((json.get('targetPaths')))
        target_classes_ids = _parse_string_list(json.get('targetClassesIds'))
        targets = []
        for elem in target_paths:
            targets.append(ByFieldName(elem))
        for elem in target_classes_ids:
            targets.append(ByClass(elem))
        return Obfuscate(targets, obfuscationType)


class HashAlgorithm(SerializableStringEnum):
    SHA3_512 = "SHA3-512"

    @staticmethod
    def parse_error(s: str) -> DtlError:
        return DtlError(_enum_parse_error("hash algorithm", s))

    @staticmethod
    def hash_algorithm_from_str(string: str) -> Union[DtlError, 'HashAlgorithm']:
        return SerializableStringEnum.from_str(HashAlgorithm)(string)


class Encrypt(ObfuscationType):
    type = "Encrypt"

    def __init__(
            self,
            key: str,
            encryption_algorithm: EncryptionAlgorithm = EncryptionAlgorithm.AES_256
    ):
        """
        Encrypt data referenced by the target parameter of Obfuscate, with later decryption possible via the key supplied.
        The key needs to be a 32-bit base64 encoded string.

        :param encryption_algorithm: the algorithm to use, via the EncryptionAlgorithm class, defaults to AES_256
        :param key: the encryption key for later decryption
        """
        validate_AES256_key(key)
        self.encryption_algorithm = encryption_algorithm
        self.key = key

    def __repr__(self):
        return ('Encrypt('
                f'encryption_algorithm: {self.encryption_algorithm}), key: {self.key}')

    def _as_payload(self) -> dict:
        res = self._base_payload()
        res['key'] = self.key
        res['encryptionAlgorithm'] = self.encryption_algorithm.value
        return res

    def _from_payload(json: dict) -> Union[DtlError, 'Encrypt']:
        key = json.get('key')
        encryption_algorithm = EncryptionAlgorithm.encryption_algorithm_from_str(json.get('encryptionAlgorithm'))
        return Encrypt(key, encryption_algorithm)


class Decrypt(Transformation):

    type_str = "Decrypt"

    def __init__(
            self,
            key: str,
            targets: List[Union[ByClass, ByFieldName]],
            encryption_algorithm: EncryptionAlgorithm = EncryptionAlgorithm.AES_256,
    ):
        """
        Decrypt data referenced by the target parameters.
        The key needs to be a 32-bit base64 encoded string.

        :param targets: a list of references to data to target with the decryption
        :param encryption_algorithm: the algorithm used to encrypt the data, via the EncryptionAlgorithm class, defaults to AES_256
        :param key: the encryption key to use for decryption
        """
        validate_AES256_key(key)
        Transformation.__init__(self, Decrypt.type_str)
        self.key = key
        self.targets = targets
        self.encryption_algorithm = encryption_algorithm

    def __repr__(self):
        return ('Decrypt('
                f'targets: {self.targets},'
                f'encryption_algorithm: {self.encryption_algorithm})')

    def _as_payload(self) -> dict:
        base = self._base_payload()
        targetClassesIds = []
        targetPaths = []

        for elem in self.targets:
            if isinstance(elem, ByClass):
                targetClassesIds.append(str(elem.class_id))
            elif isinstance(elem, ByFieldName):
                targetPaths.append(elem.field_name)

        base['targetClassesIds'] = targetClassesIds
        base['targetPaths'] = targetPaths
        base['encryptionAlgorithm'] = self.encryption_algorithm.value
        base['key'] = self.key
        return base

    def _from_payload(json: dict) -> Union[DtlError, 'Decrypt']:
        key = json.get('key')
        target_paths = _parse_list(_parse_string_list)((json.get('targetPaths')))
        target_classes_ids = _parse_string_list(json.get('targetClassesIds'))
        encryption_algorithm = EncryptionAlgorithm.encryption_algorithm_from_str(json.get('encryptionAlgorithm'))
        targets = []
        for elem in target_paths:
            targets.append(ByFieldName(elem))
        for elem in target_classes_ids:
            targets.append(ByClass(elem))

        return Decrypt(key, targets, encryption_algorithm)


class Replace(ObfuscationType):
    type = "Replace"

    def __init__(
            self,
            replacement: str,
            left_offset: Optional[int] = None,
            right_offset: Optional[int] = None
    ):
        """
        Replace the data referenced by the target parameter of Obfuscate with a supplied string

        :param replacement: the string to use as replacement
        :param left_offset: can be set to begin the replacement at the specified character count from the beginning of the target string
        :param right_offset: can be set to stop the replacement at the specified character count from the end of the target string
        """
        self.replacement = replacement
        self.left_offset = left_offset
        self.right_offset = right_offset

    def __repr__(self):
        return ('Replace('
                f'replacement: {self.replacement}, '
                f'left_offset: {self.left_offset!r}, '
                f'right_offset: {self.right_offset!r})')

    def _as_payload(self) -> dict:
        base = self._base_payload()
        base['replacement'] = self.replacement
        base['leftOffset'] = self.left_offset
        base['rightOffset'] = self.right_offset
        return base

    def _from_payload(json: dict) -> Union[DtlError, 'Replace']:
        replacement = json.get('replacement')
        leftOffset = json.get('leftOffset')
        rightOffset = json.get('rightOffset')
        return Replace(replacement, leftOffset, rightOffset)


class Hash(ObfuscationType):
    type = "Hash"

    def __init__(
            self,
            hash_algorithm: HashAlgorithm = HashAlgorithm.SHA3_512,
            salt_hash: Optional[str] = None
    ):
        """
          Hash the data referenced by the target parameter of Obfuscate

          :param hash_algorithm: the algorithm to use, via the HashAlgorithm class, defaults to SHA3-512
          :param salt_hash: can be used to provide a salt the hash, appending a string to the pre-hashed string and increasing the output’s complexity
        """
        self.hash_algorithm = hash_algorithm
        self.salt_hash = salt_hash

    def __repr__(self):
        return ('Hash('
                f'hash_algorithm: {self.hash_algorithm}, '
                f'salt_hash: {self.salt_hash})')

    def _as_payload(self) -> dict:
        base = self._base_payload()
        base['hashAlgorithm'] = self.hash_algorithm.value
        base['salt'] = self.salt_hash
        return base

    def _from_payload(json: dict) -> Union[DtlError, 'Hash']:
        salt = json.get('salt')
        hash_algo = HashAlgorithm.hash_algorithm_from_str(json.get('hash_algorithm'))
        return Hash(hash_algo, salt)

def validate_AES256_key(key: str):
    try:
        if len(b64decode(key.encode('utf-8'))) != 32:
            raise DtlError('Invalid key. Encryption/Decryption key needs to be 256-bit in length')
    except:
        raise DtlError('Invalid key. Encryption/Decryption key needs to be 256-bit in length')





from enum import Enum
from typing import Optional, List, Union

from datalogue.errors import DtlError, _enum_parse_error
from datalogue.dtl_utils import SerializableStringEnum, exists, _parse_datetime, _parse_uuid, _parse_list, map_option
from uuid import UUID
from datetime import datetime


class SearchOrder(Enum):
    created_at_asc = ("CreatedAt", "Asc")
    created_at_desc = ("CreatedAt", "Desc")
    updated_at_desc = ("UpdatedAt", "Desc")
    updated_at_asc = ("UpdatedAt", "Asc")


class RegexStatus(SerializableStringEnum):
    NoMatch = "NoMatch"
    PerfectMatch = "PerfectMatch"

    @staticmethod
    def parse_error(s: str) -> str:
        return DtlError(_enum_parse_error("regex status", s))

    @staticmethod
    def from_str(string: str) -> Union[DtlError, 'RegexStatus']:
        return SerializableStringEnum.from_str(RegexStatus)(string)


class RegexTestSample:
    """
    Representation for test samples for Regex. It contains a text value and 
    optionally a status to denote whether it's a Match or NoMatch
    """

    def __init__(self, text: str, status: Optional[RegexStatus] = None):
        self.text = text
        self.status = status

    def __repr__(self):
        return ('\n  RegexTestSample('
                f'text: {self.text!r}, '
                f'status: {self.status})')

    def __eq__(self, other):
        return (self.text == other.text) and \
               (self.status == other.status)

    @staticmethod
    def _as_payload(sample: 'RegexTestSample') -> Union[DtlError, dict]:
        payload = {"text": sample.text}
        if sample.status is not None:
            payload["status"] = sample.status.value
        return payload

    @staticmethod
    def _from_payload(payload: dict) -> Union[DtlError, 'Regex']:
        text = payload.get("text")
        if text is None:
            return DtlError("'text' is missing! Invalid json for RegexTestSample")
        status = map_option(payload.get("status"), RegexStatus.from_str)
        return RegexTestSample(text, status)


class Regex:
    """
    Representation for regexes.
    """

    def __init__(
            self,
            name: str,
            pattern: str,
            description: str = None,
            test_data: Optional[Union[List[str], List[RegexTestSample]]] = [],
            id: Optional[UUID] = None,
            owner: Optional[UUID] = None,
            editors: Optional[List[UUID]] = None,
            created_at: Optional[datetime] = None,
            updated_at: Optional[datetime] = None,
            updated_by: Optional[UUID] = None):
        self.id = id
        self.name = name
        self.pattern = pattern
        self.description = description
        if test_data is not None and all(isinstance(test_datum_text, str) for test_datum_text in test_data):
            self.test_data = list(map(lambda test_datum: RegexTestSample(test_datum), test_data))
        elif test_data is not None and all(
                isinstance(test_datum_text, RegexTestSample) for test_datum_text in test_data):
            self.test_data = test_data
        elif test_data is None:
            self.test_data = test_data
        else:
            raise DtlError("Please provide either list of strings or list of RegexTestSample for test_data parameters.")
        self.owner = owner
        self.editors = editors
        self.created_at = created_at
        self.updated_at = updated_at
        self.updated_by = updated_by

    def __eq__(self, other):
        return (self.name == other.name) and \
               (self.pattern == other.pattern) and \
               (self.description == other.description) and \
               (self.test_data == other.test_data) and \
               (self.owner == other.owner) and \
               (self.editors == other.editors) and \
               (self.created_at == other.created_at) and \
               (self.updated_at == other.updated_at) and \
               (self.updated_by == other.updated_by)

    def __repr__(self):
        return ('Regex('
                f'id: {self.id!r}, '
                f'name: {self.name!r}, '
                f'pattern: {self.pattern!r},  '
                f'description: {self.description!r}, '
                f'test_data: {self.test_data})')

    @staticmethod
    def _as_payload(regex: 'Regex') -> Union[DtlError, dict]:
        payload = {
            "name": regex.name,
            "pattern": regex.pattern,
            "description": regex.description,
            "testData": list(map(lambda t: RegexTestSample._as_payload(t), regex.test_data))
        }

        if regex.id is not None:
            payload["id"] = regex.id

        if regex.owner is not None:
            payload["owner"] = regex.owner

        if regex.editors is not None:
            payload["editors"] = regex.editors

        if regex.created_at is not None:
            payload["createdAt"] = regex.created_at

        if regex.updated_at is not None:
            payload["updatedAt"] = regex.updated_at

        if regex.updated_by is not None:
            payload["lastUpdateBy"] = regex.updated_by

        return payload

    @staticmethod
    def _from_payload(payload: dict) -> Union[DtlError, 'Regex']:
        name = payload.get("name")
        if name is None or not isinstance(name, str):
            return DtlError("'name' is missing or not a string")

        pattern = payload.get("pattern")
        if pattern is None or not isinstance(pattern, str):
            return DtlError("'pattern' is missing or not a string")

        description = payload.get("description")
        if description is None or not isinstance(description, str):
            return DtlError("'description' is missing or not a string")

        test_data_jsons = payload.get("testData")
        if test_data_jsons is None or not isinstance(test_data_jsons, List):
            return DtlError("'testData' is missing or not a List")

        test_data = _parse_list(RegexTestSample._from_payload)(test_data_jsons)
        if isinstance(test_data, DtlError):
            return test_data

        id = map_option(payload.get("id"), _parse_uuid)
        owner = map_option(payload.get("owner"), _parse_uuid)

        editors = payload.get("editors")
        if editors is not None:
            editors = _parse_list(_parse_uuid)(editors)
            if isinstance(editors, DtlError):
                return DtlError(f"'editors has at least one invalid value: {payload.get('editors')}", editors)

        created_at = map_option(payload.get("createdAt"), _parse_datetime)
        if isinstance(created_at, DtlError):
            return created_at

        updated_at = map_option(payload.get("updatedAt"), _parse_datetime)
        if isinstance(updated_at, DtlError):
            return updated_at

        updated_by = map_option(payload.get("lastUpdateBy"), _parse_uuid)
        if isinstance(updated_by, DtlError):
            return updated_by

        return Regex(name, pattern, description, test_data, id, owner, editors, created_at, updated_at, updated_by)



